import math
import torch
import torch.nn as nn
import torch.nn.functional as F


class FixupCNN(nn.Module):
    """
    A larger version of the IMPALA CNN with Fixup init.
    See Fixup: https://arxiv.org/abs/1901.09321.

    Parameters
    ----------
    input_shape : tuple
        Shape input tensors.

    Attributes
    ----------
    feature_extractor : nn.Module
        Neural network feature extractor block.
    """
    def __init__(self, input_shape):
        super(FixupCNN, self).__init__()

        depth_in = input_shape[0]
        layers = []
        for depth_out in [16, 32, 32]:
            layers.extend([
                nn.Conv2d(depth_in, depth_out, 3, padding=1),
                nn.MaxPool2d(3, stride=2, padding=1),
                FixupResidualModule(depth_out, 6),
                FixupResidualModule(depth_out, 6),
            ])
            depth_in = depth_out
        self.feature_extractor = nn.Sequential(*layers)
        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        out = self.feature_extractor(inputs / 255.0)
        return out

class FixupResidualModule(nn.Module):
    """
    FixupCNN resudial block.

    Parameters
    ----------
    depth : int
        Number of input feature maps.
    num_residual : int
        Number of residual feature blocks in the architecture
    """
    def __init__(self, depth, num_residual):
        super(FixupResidualModule, self).__init__()

        self.conv1 = nn.Conv2d(depth, depth, 3, padding=1, bias=False)
        self.conv2 = nn.Conv2d(depth, depth, 3, padding=1, bias=False)

        for p in self.conv1.parameters():
            p.data.mul_(1 / math.sqrt(num_residual))
        for p in self.conv2.parameters():
            p.data.zero_()

        self.bias1 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias2 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias3 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias4 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.scale = nn.Parameter(torch.ones([depth, 1, 1]))

    def forward(self, x):
        """
        Forward pass residual block

        Parameters
        ----------
        inputs : torch.tensor
            Input feature map.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        x = F.relu(x)
        out = x + self.bias1
        out = self.conv1(out)
        out = out + self.bias2
        out = F.relu(out)
        out = out + self.bias3
        out = self.conv2(out)
        out = out * self.scale
        out = out + self.bias4
        return out + x


import torch
from .on_policy_buffer import OnPolicyBuffer as B


class OnPolicyVTraceBuffer(B):
    """
    Storage class for On-Policy algorithms with off-policy correction method
    V-trace (https://arxiv.org/abs/1506.02438).

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.

    Attributes
    ----------
    on_policy_data_fields : tuple
        Accepted data fields. If the samples inserted contain other fields,
        an AssertionError will be raised.
    max_size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place.
    """

    on_policy_data_fields = ("obs", "obs2", "rhs", "act", "rew", "val", "logp", "done")

    def __init__(self, size, device=torch.device("cpu")):

        super(OnPolicyVTraceBuffer, self).__init__(
            size=size,
            device=device)

    def before_update(self, actor, algo):
        """
        Before updating actor policy model, compute returns and advantages.

        Parameters
        ----------
        actor : ActorCritic
            An actor class instance.
        algo : Algo
            An algorithm class instance.
        """
        self.compute_vtrace(self, actor, algo)

    @torch.no_grad()
    def get_updated_action_log_probs(self, actor, algo):
        """
        Computes new log probabilities of actions stored in `storage`
        according to current `actor` version. It also uses the current
        `actor` version to update the value predictions.

        Parameters
        ----------
        actor : ActorCritic
            An actor class instance.
        algo : Algo
            An algorithm class instance.

        Returns
        -------
        new_logp : torch.tensor
            New action log probabilities.
        """

        len, num_proc = self.data["act"].shape[0:2]

        # Create batches without shuffling data
        batches = self.generate_batches(
            algo.num_mini_batch, algo.mini_batch_size,
            num_epochs=1, recurrent_ac=actor.is_recurrent, shuffle=False)

        # Obtain new value and log probability predictions
        new_val = []
        new_logp = []
        for batch in batches:
            obs, rhs, act, done = batch["obs"], batch["rhs"], batch["act"], batch["done"]
            (val, logp, _, _) = actor.evaluate_actions(obs, rhs, done, act)
            new_val.append(val)
            new_logp.append(logp)

        # Concatenate results
        if actor.is_recurrent:
            new_val = [p.view(len, num_proc // algo.num_mini_batch, -1) for p in new_val]
            self.data["val"][:-1] = torch.cat(new_val, dim=1)
            new_logp = [p.view(len, num_proc // algo.num_mini_batch, -1) for p in new_logp]
            new_logp = torch.cat(new_logp, dim=1)
        else:
            self.data["val"][:-1] = torch.cat(new_val, dim=0).view(len, num_proc, 1)
            new_logp = torch.cat(new_logp, dim=0).view(len, num_proc, 1)

        return new_logp

    @torch.no_grad()
    def compute_vtrace(self, new_policy, algo, clip_rho_thres=1.0, clip_c_thres=1.0):
        """
        Computes V-trace target values and advantage predictions and stores them,
        along with the updated action log probabilities, in `storage`.

        Parameters
        ----------
        new_policy : Actor
            An actor class instance.
        algo : Algo
            An algorithm class instance.
        """

        l = self.step if self.step != 0 else self.max_size

        with torch.no_grad():
            _ = new_policy.get_action(
                self.data["obs"][self.step - 1],
                self.data["rhs"][self.step - 1],
                self.data["done"][self.step - 1])
            next_value = new_policy.get_value(self.data["obs"][self.step - 1])
        self.data["val"][self.step] = next_value

        new_action_log_probs = self.get_updated_action_log_probs(new_policy, algo)

        log_rhos = (new_action_log_probs - self.data["logp"])
        clipped_rhos = torch.clamp(torch.exp(log_rhos), max=clip_rho_thres)
        clipped_cs = torch.clamp(torch.exp(log_rhos), max=clip_c_thres)

        deltas = clipped_rhos * (
            self.data["rew"] + algo.gamma * self.data["val"][1:] - self.data["val"][:-1])

        acc = torch.zeros_like(self.data["val"][-1])
        result = []
        for i in reversed(range(l)):
            acc = deltas[i] + algo.gamma * clipped_cs[i] * acc * (1 - self.data["done"][i + 1])
            result.append(acc)

        result.reverse()
        result.append(torch.zeros_like(self.data["val"][-1]))
        vs_minus_v_xs = torch.stack(result)

        vs = torch.add(vs_minus_v_xs, self.data["val"])
        adv = clipped_rhos * (self.data["rew"] + algo.gamma * vs[1:] - self.data["val"][:-1])

        self.data["ret"] = vs
        self.data["logp"] = new_action_log_probs
        self.data["adv"] = (adv - adv.mean()) / (adv.std() + 1e-8)





from .gu_worker2 import GUWorker
from .c_workers import CWorkerSet

class Workers:
    """
    Class to containing and handling all scheme workers.

    Parameters
    ----------
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    col_worker_remote_config : dict
        Ray resource specs for the remote collection workers.
    num_col_workers : int
        Number of remote workers performing collection operations.
    updater_device : torch.device
        CPU or specific GPU to use for local computation.
    broadcast_interval : int
        After how many central updates, model weights should be broadcasted to
        remote collection workers.
    updater_queue_size : int
        Maximum number of data dicts fitting in the updater queue.
    max_collect_requests_pending : int
        Maximum number of collection tasks simultaneously scheduled to each
        collection worker.
    """
    def __init__(self,
                 num_col_workers,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 broadcast_interval=1,
                 updater_queue_size=100,
                 updater_device="cpu",
                 max_collect_requests_pending=2,
                 test_envs_factory=lambda x, y, c: None,
                 col_worker_remote_config={"num_cpus": 1, "num_gpus": 0.5}):

        col_workers = CWorkerSet(
            num_workers=num_col_workers,
            algo_factory=algo_factory,
            storage_factory=storage_factory,
            test_envs_factory=test_envs_factory,
            train_envs_factory=train_envs_factory,
            actor_factory=actor_factory,
            worker_remote_config=col_worker_remote_config)
        self._update_worker = GUWorker(
            col_workers,
            device=updater_device,
            broadcast_interval=broadcast_interval,
            updater_queue_size=updater_queue_size,
            max_collect_requests_pending=max_collect_requests_pending)

    def update_workers(self):
        """Return local worker"""
        return self._update_worker


from abc import ABC, abstractmethod


class Storage(ABC):
    """
    Base class for all storage components. It should serve as a template to
    create new Storage classes with new or extended features.
    """

    @classmethod
    @abstractmethod
    def create_factory(cls, *args):
        """Returns a function to create new Storage instances"""
        raise NotImplementedError

    @abstractmethod
    def init_tensors(self, sample, *args):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        raise NotImplementedError

    @abstractmethod
    def get_data(self, *args):
        """Return currently stored data."""
        raise NotImplementedError

    @abstractmethod
    def add_data(self, new_data, *args):
        """
        Replace currently stored data.

        Parameters
        ----------
        new_data : dict
            Dictionary of env transition samples to replace self.data with.
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self, *args):
        """Set class counters to zero and remove stored data"""
        raise NotImplementedError

    @abstractmethod
    def insert(self, sample, *args):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        raise NotImplementedError

    @abstractmethod
    def before_update(self, actor_critic, algo, *args):
        """
        Before updating actor policy model, compute returns and advantages.

        Parameters
        ----------
        actor_critic : ActorCritic
            An actor_critic class instance.
        algo : an algorithm class
            An algorithm class instance.
        """
        raise NotImplementedError

    @abstractmethod
    def after_update(self, *args):
        """After updating actor policy model, make sure self.step is at 0."""
        raise NotImplementedError

    @abstractmethod
    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1, recurrent_ac=False, *args):
        """
        Returns a batch iterator to update actor critic.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.
        recurrent_ac : bool
            Whether actor critic policy is a RNN or not.
        shuffle : bool
            Whether to shuffle collected data or generate sequential

        Yields
        ______
        batch : dict
            Generated data batches.
        """
        raise NotImplementedError




def get_scheme_workers(scheme_name):
    """Return scheme workers from distributed scheme name"""

    if scheme_name == "3cs":
        from .scheme_3cs import Worker
        return Worker
    elif scheme_name == "3ds":
        from .scheme_3ds import Workers
        return Workers
    elif scheme_name == "2dacs":
        from .scheme_2dacs import Workers
        return Workers
    elif scheme_name == "2daca":
        from .scheme_2daca import Workers
        return Workers
    elif scheme_name == "da2cs":
        from .scheme_da2cs import Workers
        return Workers
    elif scheme_name == "dadacs":
        from .scheme_dadacs import Workers
        return Workers
    elif scheme_name == "dadaca":
        from .scheme_dadaca import Workers
        return Workers
    else:
        raise NotImplementedError


import torch
import torch.nn as nn
from abc import ABC, abstractmethod

from .neural_networks.feature_extractors import get_model


class Actor(nn.Module, ABC):
    """
    Actor critic class for Off-Policy algorithms.

    It contains a policy network (actor) to predict next actions and one or two
    Q networks.

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    feature_extractor_network : nn.Module
        PyTorch nn.Module to extract features in all networks.
    feature_extractor_kwargs : dict
        Keyword arguments for the feature extractor network.
    recurrent_policy : bool
        Whether to use a RNN as a policy.
    recurrent_hidden_size : int
        Policy RNN hidden size.
    create_double_q_critic : bool
        Whether to instantiate a second Q network or not.

    Attributes
    ----------
    policy_net : nn.module
        Neural network that extracts features from the input env obs.
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    dist : nn.module
        Neural network that predict a prob distribution over the action space
    q1 : nn.module
        Neural network that predicts a Q-value for a given env obs and action.
    q2 : nn.module
        A second neural network to predict a Q-value for a given env obs and action.
    scale : nn.module
        Maps actions from [space.low, space.high] range to [-1, 1] range.
    unscale : nn.module
        Maps actions from [-1, 1] range to [space.low, space.high] range.

    Examples
    --------
    """
    def __init__(self,
                 input_space,
                 action_space,
                 feature_extractor_network=get_model("MLP"),
                 feature_extractor_kwargs={},
                 recurrent_policy=False,
                 recurrent_hidden_size=512,
                 *args):

        super(Actor, self).__init__()
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def create_factory(
            cls,
            input_space,
            action_space,
            restart_model=None,
            recurrent_policy=False,
            recurrent_hidden_size=512,
            feature_extractor_kwargs={},
            feature_extractor_network=get_model("MLP"),
            *args):
        """
        Returns a function that creates actor critic instances.

        Parameters
        ----------
        input_space : gym.Space
            Environment observation space.
        action_space : gym.Space
            Environment action space.
        restart_model : str
            Path to a previously trained ActorCritic checkpoint to be loaded.
        feature_extractor_network : nn.Module
            PyTorch nn.Module used as the features extraction block in all networks.
        feature_extractor_kwargs : dict
            Keyword arguments for the feature extractor network.
        recurrent_policy : bool
            Whether to use a RNN as a policy.
        recurrent_hidden_size : int
            Policy RNN hidden size.
        create_double_q_critic : bool
            whether to instantiate a second Q network or not.

        Returns
        -------
        create_actor_critic_instance : func
            creates a new OffPolicyActorCritic class instance.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def is_recurrent(self, *args):
        """Returns True if the policy network has recurrency."""
        raise NotImplementedError

    @property
    @abstractmethod
    def recurrent_hidden_state_size(self):
        """Size of policy recurrent hidden state"""
        raise NotImplementedError

    @abstractmethod
    def policy_initial_states(self, obs, *args):
        """
        Returns all policy inputs to predict the environment initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : torch.tensor
            Initial recurrent hidden state (will contain zeroes).
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """
        raise NotImplementedError

    @abstractmethod
    def get_action(self, obs, rhs, dones, deterministic=False, *args):
        """
        Predict and return next action, along with other information.

        Parameters
        ----------
        obs : torch.tensor
            Current environment observation.
        rhs : torch.tensor
            Current recurrent hidden state.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action : torch.tensor
            Next action sampled.
        clipped_action : torch.tensor
            Next action sampled, but clipped to be within the env action space.
        logp_action : torch.tensor
            Log probability of `action` within the predicted action distribution.
        rhs : torch.tensor
            Updated recurrent hidden state.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """
        raise NotImplementedError

    def evaluate_actions(self, obs, rhs, done, action):
        """
        Evaluate log likelihood of action given obs and the current
        policy network. Returns also entropy distribution.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        rhs : torch.tensor
            Recurrent hidden state.
        done : torch.tensor
            Done tensor, indicating if episode has finished.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted with current version of the policy_net.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted with current version
            of the policy_net.
        rhs : torch.tensor
            Updated recurrent hidden state.
        """
        raise NotImplementedError

    def get_q_scores(self, obs, actions=None):
        """
        If actor has Q-networks, return Q scores of the given observations
        and actions.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        actions : torch.tensor
             Evaluated actions.

        Returns
        -------
        q1 : torch.tensor
            Q score according to current q1 network version.
        q2 : torch.tensor
            Q score according to current q2 network version. If there is no
            q2 network, return None
        """
        raise NotImplementedError

    def get_value(self, obs):
        """
        if actor has V-network, return value score of given observation.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.

        Returns
        -------
        value : torch.tensor
            value score according to current value_net version.
        """
        raise NotImplementedError



import os
import torch
import inspect
from baselines import bench
from baselines.common.vec_env.dummy_vec_env import DummyVecEnv
from baselines.common.vec_env.subproc_vec_env import SubprocVecEnv

from .vector_wrappers import VecPyTorch
from .env_wrappers import TransposeImage


class VecEnv:
    """Class to handle creation of environment vectors"""

    @classmethod
    def create_factory(cls, env_fn, env_kwargs, vec_env_size, log_dir=None, info_keywords=()):
        """
        Returns a function to create a vector of environments of size
        num_processes, so it can be executed by any worker, remote or not.

        Parameters
        ----------
        env_fn : func
            Function to create the environment.
        env_kwargs : dict
            keyword arguments of env_fn.
        num_processes : int
            size of the vector of environments.
        log_dir : str
            Target path for envs to log information through bench.Monitor class.
        info_keywords : tuple
            Information keywords to be logged stored by bench.Monitor class.

        Returns
        -------
        make_vec_env : func
            Function to create a vector of environments.
        dummy_env.action_space : gym.Space
            Environments action space.
        dummy_env.observation_space: gym.Space
            Environments observation space.
        """

        def make_vec_env(device=torch.device("cpu"), index_worker=0, mode="train"):
            """Create and return a vector environment"""

            if mode == "train":
                env_indexes = range(0, vec_env_size)
            else:
                env_indexes = range(vec_env_size, 2 * vec_env_size)
            index_worker *= vec_env_size  # to avoid repeating seeds between workers

            envs = [make_env(
                env_fn=env_fn, env_kwargs=env_kwargs,
                index_worker=index_worker,
                index_env=i, log_dir=log_dir, mode=mode,
                info_keywords=info_keywords
            ) for i in env_indexes]

            if len(envs) > 1:
                envs = SubprocVecEnv(envs)
            else:
                envs = DummyVecEnv(envs)

            envs = VecPyTorch(envs, device)

            return envs

        dummy_env = [make_env(
            env_fn=env_fn, env_kwargs=env_kwargs, index_worker=0, index_env=0)]
        dummy_env = DummyVecEnv(dummy_env)

        cls.action_space = dummy_env.action_space
        cls.observation_space = dummy_env.observation_space

        return make_vec_env, dummy_env.action_space, dummy_env.observation_space

def make_env(env_fn, env_kwargs, index_worker, index_env, log_dir=None, info_keywords=(), mode="train"):
    """
    Returns a function that handles the creating of a single environment, so it
    can be executed in an independent thread.

    Parameters
    ----------

    env_fn : func
        Function to create the environment.
    env_kwargs : dict
        keyword arguments of env_fn.
    log_dir : str
        Target path for bench.Monitor logger values.
    info_keywords : tuple
        Information keywords to be logged stored by bench.Monitor.
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    mode : str
        "train" or "test"

    Returns
    -------
    _thunk : func
        A function to create and return the environment. It also sets up the
        Monitor logging and used a TransposeImage wrapper if environment obs
        are images.

    """

    if log_dir:
        path = os.path.join(log_dir, mode)
        os.makedirs(path, exist_ok=True)

    # index_worker and index_env added as paramaters
    if "index_worker" in inspect.getfullargspec(env_fn).args:
        env_kwargs["index_worker"] = index_worker
    if "index_env" in inspect.getfullargspec(env_fn).args:
        env_kwargs["index_env"] = index_env

    def _thunk():
        """Creates and returns environment"""

        # Create single environment
        env = env_fn(**env_kwargs)

        # Monitor provided info_keywords
        if log_dir is not None:
            env = bench.Monitor(
                env, os.path.join(path, "{}_{}".format(index_worker, str(index_env))),
                allow_early_resets=True, info_keywords=info_keywords)

        # if obs are images with shape (W,H,C), transpose to (C,W,H) for PyTorch convolutions
        obs_shape = env.observation_space.shape
        if len(obs_shape) == 3:
            env = TransposeImage(env, op=[2, 0, 1])

        return env

    return _thunk



import gym
from gym.spaces.box import Box


class TransposeImage(gym.ObservationWrapper):
    """
    When environment observations are images, this wrapper allows to transpose
    the axis. It is useful when the images have shape (W,H,C), as they can be
    transposed "on the fly" to (C,W,H) for PyTorch convolutions to be applied.

    Parameters
    ----------
    env : gym.Env
        Original Gym environment, previous to applying the wrapper.
    op : list
        New axis ordering.

    Attributes
    ----------
    op : list
        New axis ordering.
    observation_space : gym.Space
        New observation space
    """

    def __init__(self, env=None, op=[2, 0, 1]):
        """Transpose observation space for images"""
        super(TransposeImage, self).__init__(env)
        assert len(op) == 3, "Error: Operation, {}, must be dim3".format(str(op))
        self.op = op
        obs_shape = self.observation_space.shape
        self.observation_space = Box(
            self.observation_space.low[0, 0, 0],
            self.observation_space.high[0, 0, 0], [
                obs_shape[self.op[0]], obs_shape[self.op[1]],
                obs_shape[self.op[2]]],
            dtype=self.observation_space.dtype)

    def observation(self, ob):
        """Transpose observation"""
        return ob.transpose(self.op[0], self.op[1], self.op[2])


import torch
import torch.nn as nn


class Scale(nn.Module):
    """
    Maps inputs from [space.low, space.high] range to [-1, 1] range.

    Parameters
    ----------
    space : gym.Space
        Space to map from.

    Attributes
    ----------
    low : torch.tensor
        Lower bound for unscaled Space.
    high : torch.tensor
        Upper bound for unscaled Space.
    """
    def __init__(self, space):
        super(Scale, self).__init__()
        self.register_buffer("low", torch.from_numpy(space.low))
        self.register_buffer("high", torch.from_numpy(space.high))

    def forward(self, x):
        """
        Maps x from [space.low, space.high] to [-1, 1].

        Parameters
        ----------
        x : torch.tensor
            Input to be scaled
        """
        return 2.0 * ((x - self.low) / (self.high - self.low)) - 1.0


class Unscale(nn.Module):
    """
    Maps inputs from [-1, 1] range to [space.low, space.high] range.

    Parameters
    ----------
    space : gym.Space
        Space to map from.

    Attributes
    ----------
    low : torch.tensor
        Lower bound for unscaled Space.
    high : torch.tensor
        Upper bound for unscaled Space.
    """
    def __init__(self, space):
        super(Unscale, self).__init__()
        self.register_buffer("low", torch.from_numpy(space.low))
        self.register_buffer("high", torch.from_numpy(space.high))

    def forward(self, x):
        """
        Maps x from [-1, 1] to [space.low, space.high].

        Parameters
        ----------
        x : torch.tensor
            Input to be unscaled
        """
        return self.low + (0.5 * (x + 1.0) * (self.high - self.low))



import os
import ray
import logging
from shutil import copy2
import torch

logger = logging.getLogger(__name__)

default_remote_config = {
    "num_cpus": 1,
    "num_gpus": 0.5,
}


class Worker:
    """
    Class containing common worker functionality.

    Parameters
    ----------
    index_worker : int
        Worker index.

    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    actor : nn.Module
        An actor class instance.
    """

    def __init__(self, index_worker):
        self.index_worker = index_worker
        self.actor = None # Initialize in inherited class

    @classmethod
    def as_remote(cls,
                  num_cpus=None,
                  num_gpus=None,
                  memory=None,
                  object_store_memory=None,
                  resources=None):
        """
        Creates a Worker instance as a remote ray actor.

        Parameters
        ----------
        num_cpus : int
            The quantity of CPU cores to reserve for this Worker class.
        num_gpus  : float
            The quantity of GPUs to reserve for this Worker class.
        memory : int
            The heap memory quota for this actor (in bytes).
        object_store_memory : int
            The object store memory quota for this actor (in bytes).
        resources: Dict[str, float]
            The default resources required by the actor creation task.

        Returns
        -------
        W : Worker
            A ray remote actor Worker class.
        """
        W = ray.remote(
            num_cpus=num_cpus,
            num_gpus=num_gpus,
            memory=memory,
            object_store_memory=object_store_memory,
            resources=resources)(cls)
        return W

    def print_worker_info(self):
        """Print information about this worker, including index and resources assigned"""
        s = "Created {} with worker_index {}".format(
            str(type(self).__name__), self.index_worker)
        if self.index_worker != 0:
            s += ", in machine {} using gpus {}".format(
                ray._private.services.get_node_ip_address(),
                ray.get_gpu_ids())
        logger.warning(s)

    def get_weights(self):
        """Returns current actor.state_dict() weights"""
        return {k: v.cpu() for k, v in self.actor.state_dict().items()}

    def save_model(self, fname, iter):
        """
        Save current version of actor as a torch loadable checkpoint.

        Parameters
        ----------
        fname : str
            Filename given to the checkpoint.
        iter: int
            Training iteration.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        torch.save(self.actor.state_dict(), fname + ".tmp")
        os.rename(fname + '.tmp', fname)
        save_name = fname + ".{}".format(iter)
        copy2(fname, save_name)
        return save_name

    def terminate_worker(self):
        """Terminate this ray actor"""
        ray.actor.exit_actor()

    @staticmethod
    def get_host():
        """Return node name where this Worker is being executed."""
        return os.uname()[1]



from abc import ABC, abstractmethod
import torch


class Algo(ABC):
    """Base class for all algorithms"""

    @classmethod
    @abstractmethod
    def create_factory(cls):
        """Returns a function to create new Algo instances"""
        raise NotImplementedError

    @abstractmethod
    def acting_step(self, obs, rhs, done, deterministic=False, *args):
        """
        PPO acting function.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other: dict
            Additional PPO predictions, value score and action log probability,
            which are not used in other algorithms.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_gradients(self, batch, grads_to_cpu=True, *args):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        data: dict
            data batch containing all required tensors to compute PPO loss.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info: dict
            Dict containing current PPO iteration information.
        """

        raise NotImplementedError

    @abstractmethod
    def apply_gradients(self, gradients=None, *args):
        """
        Take an optimization step, previously setting new gradients if provided.

        Parameters
        ----------
        gradients: list of tensors
            List of actor_critic gradients.
        """
        raise NotImplementedError

    @abstractmethod
    def set_weights(self, weights, *args):
        """
        Update actor critic with the given weights

        Parameters
        ----------
        weights: dict of tensors
            Dict containing actor_critic weights to be set.
        """
        raise NotImplementedError

    @abstractmethod
    def update_algo_parameter(self, parameter_name, new_parameter_value, *args):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        raise NotImplementedError



import torch
from torch.utils.data.sampler import BatchSampler, SubsetRandomSampler, SequentialSampler
from ..base import Storage as S


class OnPolicyBuffer(S):
    """
    Storage class for On-Policy algorithms.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.

    Attributes
    ----------
    on_policy_data_fields : tuple
        Accepted data fields. If the samples inserted contain other fields,
        an AssertionError will be raised.
    max_size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place.
    """

    on_policy_data_fields = ("obs", "obs2", "rhs", "act", "rew", "val", "logp", "done")

    def __init__(self, size, device=torch.device("cpu")):

        self.device = device
        self.max_size = size
        self.reset()

    @classmethod
    def create_factory(cls, size):
        """
        Returns a function that creates OnPolicyBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.

        Returns
        -------
        create_buffer_instance : func
            creates a new OnPolicyBuffer class instance.
        """
        def create_buffer_instance(device):
            """Create and return a OnPolicyBuffer instance."""
            return cls(size, device)
        return create_buffer_instance

    def init_tensors(self, sample):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        assert set(sample.keys()) == set(self.data.keys())
        self.data["obs"] = torch.zeros(self.max_size + 1, *sample["obs"].shape).to(self.device)
        self.data["val"] = torch.zeros(self.max_size + 1, *sample["val"].shape).to(self.device)
        self.data["rhs"] = torch.zeros(self.max_size + 1, *sample["rhs"].shape).to(self.device)
        self.data["act"] = torch.zeros(self.max_size, *sample["act"].shape).to(self.device)
        self.data["rew"] = torch.zeros(self.max_size , *sample["rew"].shape).to(self.device)
        self.data["ret"] = torch.zeros(self.max_size + 1, *sample["rew"].shape).to(self.device)
        self.data["done"] = torch.zeros(self.max_size + 1, *sample["done"].shape).to(self.device)
        self.data["logp"] = torch.zeros(self.max_size, *sample["logp"].shape).to(self.device)

    def get_data(self):
        """Return currently stored data."""
        return {k: v.cpu() for k, v in self.data.items() if v is not None}

    def add_data(self, new_data):
        """
        Replace currently stored data.

        Parameters
        ----------
        new_data : dict
            Dictionary of env transition samples to replace self.data with.
        """
        self.data = {k: v.to(self.device) for k, v in new_data.items()}

    def reset(self):
        """Set class counters to zero and remove stored data"""
        self.step, self.size = 0, 0
        self.data = {k: None for k in self.on_policy_data_fields} # lazy init

    def insert(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        if self.size == 0: # data tensors lazy initialization
            self.init_tensors(sample)

        self.data["obs"][self.step + 1].copy_(sample["obs2"])
        self.data["val"][self.step].copy_(sample["val"])
        self.data["rhs"][self.step + 1].copy_(sample["rhs"])
        self.data["act"][self.step].copy_(sample["act"])
        self.data["rew"][self.step].copy_(sample["rew"])
        self.data["done"][self.step + 1].copy_(sample["done"])
        self.data["logp"][self.step].copy_(sample["logp"])

        self.step = (self.step + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def before_update(self, actor, algo):
        """
        Before updating actor policy model, compute returns and advantages.

        Parameters
        ----------
        actor : Actor
            An actor class instance.
        algo : an algorithm class
            An algorithm class instance.
        """
        with torch.no_grad():
            _ = actor.get_action(
                self.data["obs"][self.step - 1],
                self.data["rhs"][self.step - 1],
                self.data["done"][self.step - 1])
            next_value = actor.get_value(self.data["obs"][self.step - 1])

        self.data["ret"][self.step] = next_value
        self.compute_returns(algo.gamma)
        self.compute_advantages()

    def after_update(self):
        """After updating actor policy model, make sure self.step is at 0."""
        self.data["obs"][0].copy_(self.data["obs"][self.step - 1])
        self.data["rhs"][0].copy_(self.data["rhs"][self.step - 1])
        self.data["done"][0].copy_(self.data["done"][self.step - 1])

        if self.step != 0:
            self.step = 0

    def compute_returns(self, gamma):
        """
        Compute return values.

        Parameters
        ----------
        gamma : float
            Algorithm discount factor parameter.
        """
        len = self.step if self.step != 0 else self.max_size
        for step in reversed(range(len)):
            self.data["ret"][step] = (self.data["ret"][step + 1] * gamma * (
                1.0 - self.data["done"][step + 1]) + self.data["rew"][step])

    def compute_advantages(self):
        """Compute transition advantage values."""
        adv = self.data["ret"][:-1] - self.data["val"][:-1]
        self.data["adv"] = (adv - adv.mean()) / (adv.std() + 1e-8)

    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1, recurrent_ac=False, shuffle=True):
        """
        Returns a batch iterator to update actor.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.
        recurrent_ac : bool
            Whether actor policy is a RNN or not.
        shuffle : bool
            Whether to shuffle collected data or generate sequential

        Yields
        ______
        batch : dict
            Generated data batches.
        """

        num_proc = self.data["obs"].shape[1]
        l = self.step if self.step != 0 else self.max_size

        if recurrent_ac:  # Batches for a feed recurrent actor
            num_envs_per_batch = num_proc // num_mini_batch
            assert num_proc >= num_mini_batch, "number processes greater than  mini batches"
            perm = torch.randperm(num_proc) if shuffle else list(range(num_proc))

            for _ in range(num_epochs):
                for start_ind in range(0, num_proc, num_envs_per_batch):
                    obs, rhs, act, val, ret, done, logp, adv = [], [], [], [], [], [], [], []

                    for offset in range(num_envs_per_batch):
                        ind = perm[start_ind + offset]
                        obs.append(self.data["obs"][:l, ind])
                        val.append(self.data["val"][:l, ind])
                        rhs.append(self.data["rhs"][0:1, ind])
                        act.append(self.data["act"][:l, ind])
                        ret.append(self.data["ret"][:l, ind])
                        done.append(self.data["done"][:l, ind])
                        logp.append(self.data["logp"][:l, ind])
                        if "adv" in self.data.keys():
                            adv.append(self.data["adv"][:l, ind])

                    batch = dict(
                        obs=torch.stack(obs, dim=1).view(-1, *self.data["obs"].size()[2:]),
                        val=torch.stack(val, dim=1).view(-1, *self.data["val"].size()[2:]),
                        rhs=torch.stack(rhs, dim=1).view(-1, *self.data["rhs"].size()[2:]),
                        act=torch.stack(act, dim=1).view(-1, *self.data["act"].size()[2:]),
                        ret=torch.stack(ret, dim=1).view(-1, *self.data["ret"].size()[2:]),
                        done=torch.stack(done, dim=1).view(-1, *self.data["done"].size()[2:]),
                        logp=torch.stack(logp, dim=1).view(-1, *self.data["logp"].size()[2:]),
                        adv=torch.stack(adv, dim=1).view(-1, *self.data["adv"].size()[2:])
                        if "adv" in self.data.keys() else None)

                    yield batch

        else:
            mini_batch_size = num_proc * l // num_mini_batch
            sampler = SubsetRandomSampler if shuffle else SequentialSampler
            for _ in range(num_epochs):
                for idxs in BatchSampler(sampler(range(num_proc * l)), mini_batch_size, drop_last=shuffle):
                    batch = dict(
                        obs=self.data["obs"][0:l].reshape(-1, *self.data["obs"].shape[2:])[idxs],
                        val=self.data["val"][0:l].reshape(-1, *self.data["val"].shape[2:])[idxs],
                        rhs=self.data["rhs"][0:l].reshape(-1, *self.data["rhs"].shape[2:])[idxs],
                        act=self.data["act"][0:l].reshape(-1, *self.data["act"].shape[2:])[idxs],
                        ret=self.data["ret"][0:l].reshape(-1, *self.data["ret"].shape[2:])[idxs],
                        done=self.data["done"][0:l].reshape(-1, *self.data["done"].shape[2:])[idxs],
                        logp=self.data["logp"][0:l].reshape(-1, *self.data["logp"].shape[2:])[idxs],
                        adv=self.data["adv"][0:l].reshape(-1, *self.data["adv"].shape[2:])[idxs]
                        if "adv" in self.data.keys() else None)

                    yield batch





from .gaussian import DiagGaussian
from .categorical import Categorical
from .squashed_gaussian import SquashedGaussian


def get_dist(name):
    """Returns model class from name."""
    if name == "Categorical":
        return Categorical
    elif name == "Gaussian":
        return DiagGaussian
    elif name == "SquashedGaussian":
        return SquashedGaussian
    else:
        raise ValueError("Specified model not found!")


import time
import torch
import numpy as np
from ..base.worker import Worker as W
from ..base.worker_set import WorkerSet as WS
from ..base.worker import default_remote_config


class CGWorker(W):
    """
     Worker class handling data collection and gradient computation.

    This class wraps an actor instance, a storage class instance and a
    train and a test vector of environments. It collects data, computes gradients
    and evaluates network versions following a logic defined in function self.step(),
    which will be called from the Learner class.

    Parameters
    ----------
    index_worker : int
        Worker index.
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    initial_weights : ray object ID
        Initial model weights.

    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    actor : nn.Module
        An actor class instance.
    algo : an algorithm class
        An algorithm class instance.
    envs_train : VecEnv
        A VecEnv class instance with the train environments.
    envs_test : VecEnv
        A VecEnv class instance with the test environments.
    storage : a rollout storage class
        A Storage class instance.
    iter : int
        Times gradients have been computed and sent.
    ac_version : int
        Number of times the current actor version been has been updated.
    update_every : int
        Number of data samples to collect between network update stages.
    obs : torch.tensor
        Latest train environment observation.
    rhs : torch.tensor
        Latest policy recurrent hidden state.
    done : torch.tensor
        Latest train environment done flag.
    """

    def __init__(self,
                 index_worker,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda x, y, c: None,
                 initial_weights=None):

        super(CGWorker, self).__init__(index_worker)

        # Using Ray, worker should only see one GPU or None
        device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        # Create Actor Critic instance
        self.actor = actor_factory(device)
        self.actor.to(device)

        # Create Algorithm instance
        self.algo = algo_factory(self.actor, device)

        # Define counters and other attributes
        self.iter = 0
        self.ac_version = 0
        self.num_updates = self.algo.num_epochs * self.algo.num_mini_batch

        if initial_weights: # if remote worker

            # Create train environments, define initial train states
            self.envs_train = train_envs_factory(device, index_worker)
            self.obs, self.rhs, self.done = self.actor.policy_initial_states(self.envs_train.reset())

            # Create test environments (if creation function available)
            self.envs_test = test_envs_factory(device, index_worker, mode="test")

            # Create Storage instance and set world initial state
            self.storage = storage_factory(device)
            self.update_every = self.algo.update_every or self.storage.max_size

            # Print worker information
            self.print_worker_info()

            # Set initial weights
            self.set_weights(initial_weights)

            # Collect initial samples
            print("Collecting initial samples...")
            _, self.start_steps = self.collect_data(self.algo.start_steps)

    def collect_data(self, num_steps):
        """
        Collect data from interactions with the environments.

        Parameters
        ----------
        num_steps : int
            Target number of train environment steps to take.

        Returns
        -------
        total_time : float
            Elapsed time in seconds.
        total_samples : int
            Total num of collected samples.
        """

        t = time.time()
        for step in range(num_steps):

            # Predict next action, next rnn hidden state and algo-specific outputs
            act, clip_act, rhs, algo_data = self.algo.acting_step(self.obs, self.rhs, self.done)

            # Interact with envs_vector with predicted action (clipped within action space)
            obs2, reward, done, infos = self.envs_train.step(clip_act)

            # Prepare transition dict
            transition = {"obs": self.obs, "rhs": rhs, "act": act, "rew": reward, "obs2": obs2, "done": done}
            transition.update(algo_data)

            # Store transition in buffer
            self.storage.insert(transition)

            # Update current world state
            self.obs, self.rhs, self.done = obs2, rhs, done

        # Record model version used to collect data
        self.storage.ac_version = self.ac_version

        # Record and return metrics
        total_time = time.time() - t
        total_samples = num_steps * self.envs_train.num_envs

        return total_time, total_samples

    def compute_gradients(self, batch):
        """
        Calculate actor gradients.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute algo loss.

        Returns
        -------
        grads: list of tensors
            List of actor gradients.
        info : dict
            Summary dict of relevant gradient-related information.
        """

        t = time.time()
        grads, info = self.algo.compute_gradients(batch)
        info.update({"scheme/seconds_to/compute_grads": time.time() - t})

        return grads, info

    def update_networks(self, gradients):
        """Update Actor Critic model"""
        self.algo.apply_gradients(gradients)

    def evaluate(self):
        """
        Test current actor version in self.envs_test.

        Returns
        -------
        mean_test_perf : float
            Average accumulated reward over all tested episodes.
        """

        completed_episodes = []
        obs = self.envs_test.reset()
        rewards = np.zeros(obs.shape[0])
        obs, rhs, done = self.actor.policy_initial_states(obs)

        while len(completed_episodes) < self.algo.num_test_episodes:

            # Predict next action and rnn hidden state
            act, clip_act, rhs, _ = self.algo.acting_step(obs, rhs, done, deterministic=True)

            # Interact with env with predicted action (clipped within action space)
            obs2, reward, done, _ = self.envs_test.step(clip_act)

            # Keep track of episode rewards and completed episodes
            rewards += reward.cpu().squeeze(-1).numpy()
            completed_episodes.extend(rewards[done.cpu().squeeze(-1).numpy() == 1.0].tolist())
            rewards[done.cpu().squeeze(-1).numpy() == 1.0] = 0.0

            obs = obs2

        return np.mean(completed_episodes)

    def step(self):
        """
        Perform logical learning step. Training proceeds interleaving collection
        of data samples with policy gradient computations.

        Returns
        -------
        grads: list of tensors
            List of actor gradients.
        info : dict
            Summary dict with relevant step information.
        """

        # Collect data and prepare data batches
        if self.iter % self.num_updates == 0:
            collect_time, collected_samples = self.collect_data(self.update_every)
            self.storage.before_update(self.actor, self.algo)
            self.batches = self.storage.generate_batches(
                self.algo.num_mini_batch, self.algo.mini_batch_size,
                self.algo.num_epochs, self.actor.is_recurrent)
        else:
            collect_time, collected_samples = 0.0, 0

        # Compute gradients, get algo info
        grads, info = self.compute_gradients(self.batches.__next__())

        # Add extra information to info dict
        info.update({"ac_update_num": self.iter})
        info.update({"collected_samples": collected_samples})
        info.update({"scheme/seconds_to/collect": collect_time})
        info.update({"scheme/metrics/collection_gradient_delay": self.iter - self.storage.ac_version})
        if self.iter == 0: info["collected_samples"] += self.start_steps

        # Evaluate current network
        if self.iter % self.algo.test_every == 0:
            if self.envs_test and self.algo.num_test_episodes > 0:
                test_perf = self.evaluate()
                info.update({"scheme/metrics/test_performance": test_perf})

        # Update counter
        self.iter += 1
        
        return grads, info

    def set_weights(self, weights):
        """
        Update the worker actor version with provided weights.

        weights: dict of tensors
            Dict containing actor weights to be set.
        """
        self.ac_version = weights["update"]
        self.algo.set_weights(weights["weights"])

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.algo.update_algo_parameter(parameter_name, new_parameter_value)

class CGWorkerSet(WS):
    """
    Class to better handle the operations of ensembles of CGWorkers.

    Parameters
    ----------
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    num_workers : int
        Number of remote workers in the worker set.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword arguments of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    """

    def __init__(self,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda x, y, c: None,
                 worker_remote_config=default_remote_config,
                 num_workers=1):

        self.worker_class = CGWorker
        default_remote_config.update(worker_remote_config)
        self.remote_config = default_remote_config
        self.worker_params = {
            "algo_factory": algo_factory,
            "storage_factory": storage_factory,
            "test_envs_factory": test_envs_factory,
            "train_envs_factory": train_envs_factory,
            "actor_factory": actor_factory,
        }

        super(CGWorkerSet, self).__init__(
            worker=self.worker_class,
            worker_params=self.worker_params,
            worker_remote_config=self.remote_config,
            num_workers=num_workers)


from .u_worker import UWorker
from .cg_workers import CGWorkerSet


class Workers:
    """
    Class to containing and handling all scheme workers.

    Parameters
    ----------
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    num_cg_workers : int
        Number of remote workers performing collection/gradient computation
        operations.
    """
    def __init__(self,
                 num_col_grad_workers,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda x, y, c: None,
                 worker_remote_config={"num_cpus": 1, "num_gpus": 0.5}):

        col_grad_workers = CGWorkerSet(
            num_workers=num_col_grad_workers,
            algo_factory=algo_factory,
            actor_factory=actor_factory,
            storage_factory=storage_factory,
            test_envs_factory=test_envs_factory,
            train_envs_factory=train_envs_factory,
            worker_remote_config=worker_remote_config)
        self._update_worker = UWorker(grad_workers=col_grad_workers)

    def update_workers(self):
        """Return local worker"""
        return self._update_worker


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.normal import Normal
from ..neural_networks.feature_extractors.utils import init


class SquashedGaussian(nn.Module):
    """
    Squashed Gaussian action distribution.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of action in action space.
    predict_log_std : bool
        Whether to use a nn.linerar layer to predict the output actions std.

    Attributes
    ----------
    mean: nn.Module
        Maps the incoming feature maps to output action mean values.
    log_std : nn.Module or nn.Parameter
        If predict_log_std is True Maps the incoming feature maps to output
        action std values. Otherwise, the std values are a learnable nn.Parameter.
    LOG_STD_MAX : float
        Maximum std allowed value. Used for clipping.
    LOG_STD_MIN : float
        Minimum std allowed value. Used for clipping.
    """

    LOG_STD_MAX = 2
    LOG_STD_MIN = -20

    def __init__(self, num_inputs, num_outputs, predict_log_std=True):
        super(SquashedGaussian, self).__init__()

        self.predict_log_std = predict_log_std
        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))
        self.mean = init_(nn.Linear(num_inputs, num_outputs))

        if predict_log_std:
            self.log_std = init_(nn.Linear(num_inputs, num_outputs))
        else:
            self.log_std = nn.Parameter(torch.zeros(num_outputs).unsqueeze(0))

    def forward(self, x, deterministic=False):
        """
        Predict distribution parameters from x (obs features) and return
        actions (detemrninistic, sampled and clipped), sampled action log
        probability and distribution entropy.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        logp_action : torch.tensor
            Log probability of `action` according to the predicted action distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """

        # Predict distribution parameters
        action_mean = self.mean(x)
        action_logstd = self.log_std(x) if self.predict_log_std else\
            torch.zeros(action_mean.size()).to(x.device) + self.log_std
        action_logstd = torch.clamp(action_logstd, self.LOG_STD_MIN, self.LOG_STD_MAX)

        # Create distribution and sample action
        action_dist = Normal(action_mean, action_logstd.exp())

        if deterministic:
            action = action_mean
        else:
            action = action_dist.rsample()

        # Action log probability
        logp_action = action_dist.log_prob(action).sum(axis=-1, keepdim=True)
        logp_action -= (2 * (np.log(2) - action - F.softplus(-2 * action))).sum(
            axis=1, keepdim=True)

        # Apply squashing
        action = clipped_action = torch.tanh(action)

        # Distribution entropy
        entropy_dist = -logp_action.mean()

        return action, clipped_action, logp_action, entropy_dist



import os
import ray
import time
import torch
from shutil import copy2
from collections import defaultdict
from ..utils import ray_get_and_free, average_gradients


class UWorker:
    """
    Update worker. Handles actor updates.

    This worker receives gradients from gradient workers and then updates the
    its actor model. Updated weights are synchronously sent back
    to gradient workers.

    Parameters
    ----------
    grad_workers : WorkerSet
        Set of workers computing and sending gradients to the UWorker.

    Attributes
    ----------
    num_updates : int
        Number of times the actor model has been updated.
    grad_workers : WorkerSet
        Set of workers computing and sending gradients to the UWorker.
    num_workers : int
        number of remote workers computing gradients.
    """

    def __init__(self, grad_workers):

        self.num_updates = 0
        self.grad_workers = grad_workers
        self.num_workers = len(self.grad_workers.remote_workers())

        # Check remote workers exist
        if len(self.grad_workers.remote_workers()) == 0:
            raise ValueError("""at least 1 grad worker required""")

    def sync_weights(self):
        """Synchronize gradient worker models with updater worker model"""
        weights = ray.put({
            "update": self.num_updates,
            "weights": self.grad_workers.local_worker().get_weights()})
        for e in self.grad_workers.remote_workers():
            e.set_weights.remote(weights)

    def step(self):
        """
        Takes a logical asynchronous optimization step.

        Returns
        -------
        info : dict
            Summary dict of relevant information about the update process.
        """

        to_average = []
        pending_gradients = {}
        step_metrics = defaultdict(float)

        # Call for gradients from all workers
        for e in self.grad_workers.remote_workers():
            future = e.step.remote()
            pending_gradients[future] = e

        # Wait for workers to send back gradients
        while pending_gradients:

            # Get gradients 1 by 1
            wait_results = ray.wait(list(pending_gradients.keys()))
            ready_list = wait_results[0]
            future = ready_list[0]
            gradients, info = ray_get_and_free(future)
            pending_gradients.pop(future)

            # Update info dict
            info["scheme/metrics/gradient_update_delay"] = self.num_updates - info.pop("ac_update_num")

            # Update counters
            for k, v in info.items(): step_metrics[k] += v

            # Store gradients to average later
            to_average.append(gradients)

        # Average and apply gradients
        t = time.time()
        self.grad_workers.local_worker().update_networks(average_gradients(to_average))
        avg_grads_t = time.time() - t

        # Update workers with current weights
        t = time.time()
        self.sync_weights()
        sync_grads_t = time.time() - t

        # Update counter
        self.num_updates += 1

        # Update info dict
        info = {k: v/self.num_workers if k != "collected_samples" else
        v for k, v in step_metrics.items()}
        info.update({"scheme/seconds_to/avg_grads": avg_grads_t})
        info.update({"scheme/seconds_to/sync_grads": sync_grads_t})

        return info

    def save_model(self, fname):
        """
        Save current version of actor as a torch loadable checkpoint.

        Parameters
        ----------
        fname : str
            Filename given to the checkpoint.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        torch.save(self.grad_workers.local_worker().actor.state_dict(), fname + ".tmp")
        os.rename(fname + '.tmp', fname)
        save_name = fname + ".{}".format(self.num_updates)
        copy2(fname, save_name)
        return save_name

    def stop(self):
        """Stop remote workers"""
        for e in self.grad_workers.remote_workers():
            e.terminate_worker.remote()

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.grad_workers.local_worker().update_algo_parameter(parameter_name, new_parameter_value)
        for e in self.grad_workers.remote_workers():
            e.update_algo_parameter.remote(parameter_name, new_parameter_value)


import torch
import itertools
import torch.nn as nn
import torch.optim as optim
from .base import Algo


class PPO(Algo):
    """
    Proximal Policy Optimization algorithm class.

    Algorithm class to execute PPO, from Schulman et al.
    (https://arxiv.org/abs/1707.06347). Algorithms are modules generally
    required by multiple workers, so PPO.algo_factory(...) returns a function
    that can be passed on to workers to instantiate their own PPO module.

    Parameters
    ----------
    device: torch.device
        CPU or specific GPU where class computations will take place.
    actor_critic : ActorCritic
        Actor_critic class instance.
    lr : float
        Optimizer learning rate.
    eps : float
        Optimizer epsilon parameter.
    num_epochs : int
        Number of PPO epochs.
    gamma : float
        Discount factor parameter.
    clip_param : float
        PPO clipping parameter.
    num_mini_batch : int
        Number of batches to create from collected data for actor_critic updates.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations.
    max_grad_norm : float
        Gradient clipping parameter.
    entropy_coef : float
        PPO entropy coefficient parameter.
    value_loss_coef : float
        PPO value coefficient parameter.
    use_clipped_value_loss : bool
        Prevent value loss from shifting too fast.

    Attributes
    ----------
    start_steps : int
        Number of steps collected with initial random policy (default to 0 for
        On-policy algos).
    num_epochs : int
        Times data in the buffer is re-used before data collection proceeds.
    update_every : int
        Number of data samples collected between network update stages (depends
        on storage capacity for On-policy algos).
    num_mini_batch : int
        Number mini batches per epoch.
    mini_batch_size : int
        Size of update mini batches.
    test_every : int
        Number of network updates between test evaluations.
    num_test_episodes : int
        Num episodes to complete in each test phase.
    device : torch.device
        CPU or specific GPU where class computations will take place.
    actor_critic : ActorCritic
        ActorCritic Class containing Neural Network function approximators.
    gamma : float
        Discount factor parameter.
    clip_param : float
        PPO clipping parameter.
    entropy_coef : float
        PPO entropy coefficient parameter.
    max_grad_norm : float
        Gradient clipping parameter.
    value_loss_coef : float
        PPO value coefficient parameter.
    use_clipped_value_loss : bool
        Whether or not value loss clipping is being used
    optimizer : torch.optimizer
        ActorCritic model optimizer

    Examples
    --------
    >>> ac = ActorCritic(...)
    >>> create_algo = PPO.algo_factory(
        lr=0.01, eps=1e-5, num_epochs=4, clip_param=0.2,
        entropy_coef=0.01, value_loss_coef=0.5, max_grad_norm=0.5,
        num_mini_batch=4, use_clipped_value_loss=True, gamma=0.99)
    >>> ppo1 = create_algo(actor_critic=ac, device=torch.device("cuda:1"))
    >>> ppo2 = create_algo(actor_critic=ac, device=torch.device("cuda:2"))

    >>> ac = ActorCritic(...)
    >>> ppo = PPO(device=torch.device("cuda:1"), actor_critic=ac, lr=0.01,
                  eps=1e-5, num_epochs=4, clip_param=0.2, entropy_coef=0.01,
                  value_loss_coef=0.5, max_grad_norm=0.5, num_mini_batch=4,
                  use_clipped_value_loss=True, gamma=0.99)
    """

    def __init__(self,
                 device,
                 actor_critic,
                 lr=1e-4,
                 eps=1e-8,
                 gamma=0.99,
                 num_epochs=4,
                 clip_param=0.2,
                 num_mini_batch=1,
                 test_every=20000,
                 max_grad_norm=0.5,
                 entropy_coef=0.01,
                 value_loss_coef=0.5,
                 num_test_episodes=0,
                 use_clipped_value_loss=True):

        # ---- General algo attributes ----------------------------------------

        # Number of steps collected with initial random policy
        self.start_steps = 0  # Default to 0 for On-policy algos

        # Times data in the buffer is re-used before data collection proceeds
        self.num_epochs = num_epochs

        # Number of data samples collected between network update stages
        self.update_every = None  # Depends on storage capacity

        # Number mini batches per epoch
        self.num_mini_batch = num_mini_batch

        # Size of update mini batches
        self.mini_batch_size = None  # Depends on storage capacity

        # Number of network updates between test evaluations
        self.test_every = test_every

        # Number of episodes to complete when testing
        self.num_test_episodes = num_test_episodes

        # ---- PPO-specific attributes ----------------------------------------

        self.device = device
        self.actor_critic = actor_critic
        self.gamma = gamma
        self.clip_param = clip_param
        self.entropy_coef = entropy_coef
        self.max_grad_norm = max_grad_norm
        self.value_loss_coef = value_loss_coef
        self.use_clipped_value_loss = use_clipped_value_loss

        # ----- Optimizers ----------------------------------------------------

        ac_params = itertools.chain(
            self.actor_critic.policy_net.parameters(),
            self.actor_critic.dist.parameters(),
            self.actor_critic.value_net.parameters())

        self.optimizer = optim.Adam(ac_params, lr=lr, eps=eps)

    @classmethod
    def create_factory(cls,
                     lr=1e-4,
                     eps=1e-8,
                     gamma=0.99,
                     num_epochs=4,
                     clip_param=0.2,
                     num_mini_batch=1,
                     test_every=20000,
                     max_grad_norm=0.5,
                     entropy_coef=0.01,
                     value_loss_coef=0.5,
                     num_test_episodes=0,
                     use_clipped_value_loss=True):
        """
        Returns a function to create new PPO instances.

        Parameters
        ----------
        actor_critic : ActorCritic
            Actor_critic class instance.
        lr : float
            Optimizer learning rate.
        eps : float
            Optimizer epsilon parameter.
        num_epochs : int
            Number of PPO epochs.
        gamma : float
            Discount factor parameter.
        clip_param : float
            PPO clipping parameter.
        num_mini_batch : int
            Number of batches to create from collected data for actor_critic update.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        test_every : int
            Regularity of test evaluations.
        max_grad_norm : float
            Gradient clipping parameter.
        entropy_coef : float
            PPO entropy coefficient parameter.
        value_loss_coef : float
            PPO value coefficient parameter.
        use_clipped_value_loss : bool
            Prevent value loss from shifting too fast.

        Returns
        -------
        create_algo_instance : func
            creates a new PPO class instance.
        """
        def create_algo_instance(actor_critic, device):
            return cls(lr=lr,
                       eps=eps,
                       gamma=gamma,
                       device=device,
                       test_every=test_every,
                       num_epochs=num_epochs,
                       clip_param=clip_param,
                       actor_critic=actor_critic,
                       entropy_coef=entropy_coef,
                       max_grad_norm=max_grad_norm,
                       num_mini_batch=num_mini_batch,
                       value_loss_coef=value_loss_coef,
                       num_test_episodes=num_test_episodes,
                       use_clipped_value_loss=use_clipped_value_loss)

        return create_algo_instance

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        PPO acting function.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other: dict
            Additional PPO predictions, value score and action log probability,
            which are not used in other algorithms.
        """

        with torch.no_grad():

            (action, clipped_action, logp_action, rhs,
             entropy_dist) = self.actor_critic.get_action(
                obs, rhs, done, deterministic)
            value = self.actor_critic.get_value(obs)
            other = {"val": value, "logp": logp_action}

        return action, clipped_action, rhs, other

    def compute_loss(self, data):
        """
        Compute PPO loss from data batch.

        Parameters
        ----------
        data: dict
            Data batch dict containing all required tensors to compute PPO loss.

        Returns
        -------
        value_loss: torch.tensor
            value term of PPO loss.
        action_loss: torch.tensor
            policy term of PPO loss.
        dist_entropy: torch.tensor
            policy term of PPO loss.
        loss: torch.tensor
            PPO loss.
        """

        o, rhs, a, old_v = data["obs"], data["rhs"], data["act"],  data["val"]
        r, d, old_logp, adv = data["ret"], data["done"], data["logp"], data["adv"]

        new_logp, dist_entropy, _ = self.actor_critic.evaluate_actions(o, rhs, d, a)
        new_v = self.actor_critic.get_value(o)

        ratio = torch.exp(new_logp - old_logp)
        surr1 = ratio * adv
        surr2 = torch.clamp(ratio, 1.0 - self.clip_param, 1.0 + self.clip_param) * adv
        action_loss = - torch.min(surr1, surr2).mean()

        if self.use_clipped_value_loss:
            value_losses = (new_v - r).pow(2)
            value_pred_clipped = old_v + (
            new_v - old_v).clamp(-self.clip_param, self.clip_param)
            value_losses_clipped = (value_pred_clipped - r).pow(2)
            value_loss = 0.5 * torch.max(value_losses, value_losses_clipped).mean()
        else:
            value_loss = 0.5 * (r - new_v).pow(2).mean()

        loss = value_loss * self.value_loss_coef + action_loss - self.entropy_coef * dist_entropy

        return value_loss, action_loss, dist_entropy, loss

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        data: dict
            data batch containing all required tensors to compute PPO loss.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info: dict
            Dict containing current PPO iteration information.
        """

        self.optimizer.zero_grad()
        losses = self.compute_loss(batch)
        value_loss, action_loss, dist_entropy, loss = losses
        loss.backward()
        nn.utils.clip_grad_norm_(self.actor_critic.parameters(), self.max_grad_norm)

        grads = []
        for p in self.actor_critic.parameters():
            if grads_to_cpu:
                if p.grad is not None: grads.append(p.grad.data.cpu().numpy())
                else: grads.append(None)
            else:
                if p.grad is not None:
                    grads.append(p.grad)

        info = {
            "algo/loss": loss.item(),
            "algo/value_loss": value_loss.item(),
            "algo/action_loss": action_loss.item(),
            "algo/entropy_loss": dist_entropy.item()
        }

        return grads, info

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.

        Parameters
        ----------
        gradients: list of tensors
            List of actor_critic gradients.
        """
        if gradients:
            for g, p in zip(gradients, self.actor_critic.parameters()):
                if g is not None:
                    p.grad = torch.from_numpy(g).to(self.device)
        self.optimizer.step()

    def set_weights(self, weights):
        """
        Update actor critic with the given weights

        Parameters
        ----------
        weights: dict of tensors
            Dict containing actor_critic weights to be set.
        """
        self.actor_critic.load_state_dict(weights)

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.optimizer.param_groups:
                param_group['lr'] = new_parameter_value




import torch
import torch.nn as nn
from .utils import init


class CNN(nn.Module):
    """
    Convolutional Neural Network.

    Parameters
    ----------
    input_shape : tuple
        Shape input tensors.
    activation : func
        Non-linear activation function.
    strides : list
        Convolutional layers strides.
    filters : list
        Convolutional layers number of filters.
    kernel_sizes : list
        Convolutional layers kernel sizes.

    Attributes
    ----------
    feature_extractor : nn.Module
        Neural network feature extractor block.
    """
    def __init__(self,
                 input_shape,
                 activation=nn.ReLU,
                 strides=[4, 2, 1],
                 filters=[32, 64, 64],
                 kernel_sizes=[8, 4, 3]):
        super(CNN, self).__init__()

        assert len(filters) == len(strides) and len(strides) == len(kernel_sizes)

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0), nn.init.calculate_gain('relu'))

        # Define feature extractor
        layers = []
        filters = [input_shape[0]] + filters
        for j in range(len(filters) - 1):
            layers += [init_(nn.Conv2d(
                filters[j], filters[j + 1], stride=strides[j],
                kernel_size=kernel_sizes[j])), activation()]
        self.feature_extractor = nn.Sequential(*layers)
        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        out = self.feature_extractor(inputs / 255.0)
        return out



import torch
import torch.nn as nn
from .utils import init


class MLP(nn.Module):
    """
    Multilayer Perceptron network

    Parameters
    ----------
    input_shape : tuple
        Shape input tensors.
    hidden_sizes : list
        Hidden layers sizes.
    activation : func
        Non-linear activation function.

    Attributes
    ----------
    feature_extractor : nn.Module
        Neural network feature extractor block.
    """
    def __init__(self, input_shape, hidden_sizes=[256, 256], activation=nn.ReLU):
        super(MLP, self).__init__()

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init. constant_(x, 0), nn.init.calculate_gain('relu'))

        # Define feature extractor
        layers = []
        sizes = [input_shape[0]] + hidden_sizes
        for j in range(len(sizes) - 1):
            layers += [init_(nn.Linear(sizes[j], sizes[j + 1])), activation()]
        self.feature_extractor = nn.Sequential(*layers)

        self.train()


    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        out = self.feature_extractor(inputs)
        return out



import torch
from baselines.common.vec_env import VecEnvWrapper


class VecPyTorch(VecEnvWrapper):
    """
    This wrapper turns obs, reward's and done's from numpy arrays to pytorch
    tensors and places them in the specified device, facilitating interaction
    between the environment and the actor critic function approximators (NNs).

    Parameters
    ----------
    venv : VecEnv
        Original vector environment, previous to applying the wrapper.
    device : torch.device
        CPU or specific GPU where obs, reward's and done's are placed after
        being transformed into pytorch tensors.

    Attributes
    ----------
    device : torch.device
        CPU or specific GPU where obs, reward's and done's are placed after
        being transformed into pytorch tensors.
    num_envs : int
        Size of vector environment.

    """
    def __init__(self, venv, device):
        """Return only every `skip`-th frame"""
        super(VecPyTorch, self).__init__(venv)
        self.device = device
        self.num_envs = venv.num_envs

    def reset(self):
        """New vec env reset function"""
        obs = self.venv.reset()
        obs = torch.from_numpy(obs).float().to(self.device)
        return obs

    def step_async(self, actions):
        """New vec env step_async function"""
        if isinstance(actions, torch.Tensor):
            # Squeeze the dimension for discrete actions
            actions = actions.squeeze(1)
        actions = actions.cpu().numpy()
        self.venv.step_async(actions)

    def step_wait(self):
        """New vec env step_wait function"""
        obs, reward, done, info = self.venv.step_wait()
        obs = torch.from_numpy(obs).float().to(self.device)
        reward = torch.from_numpy(reward).unsqueeze(dim=1).float().to(self.device)
        done = torch.from_numpy(done).unsqueeze(dim=1).float().to(self.device)

        return obs, reward, done, info


import os
import ray
import time
import torch
import numpy as np
from shutil import copy2
from collections import defaultdict
from ray.services import get_node_ip_address

from .utils import broadcast_message
from ..base.worker import Worker as W
from ..base.worker_set import WorkerSet as WS
from ..base.worker import default_remote_config
from .utils import check_message, find_free_port

class CGUWorker(W):
    """
    Worker class handling remote data collection, gradient computation and
    policy updates.

    This class wraps an actor instance, a storage class instance and a
    train and a test vector of environments. It collects data, computes gradients,
    updates the networks and evaluates network versions following a logic
    defined in function self.step(), which will be called from the Learner.

    Parameters
    ----------
    index_worker : int
        Worker index.
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    initial_weights : ray object ID
        Initial model weights.

    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    actor : nn.Module
        An actor class instance.
    algo : an algorithm class
        An algorithm class instance.
    envs_train : VecEnv
        A VecEnv class instance with the train environments.
    envs_test : VecEnv
        A VecEnv class instance with the test environments.
    storage : a rollout storage class
        A Storage class instance.
    iter : int
        Times actor has been updated.
    update_every : int
        Number of data samples to collect between network update stages.
    obs : torch.tensor
        Latest train environment observation.
    rhs : torch.tensor
        Latest policy recurrent hidden state.
    done : torch.tensor
        Latest train environment done flag.
    """

    def __init__(self,
                 index_worker,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda x, y, c: None,
                 initial_weights=None):

        super(CGUWorker, self).__init__(index_worker)

        # Using Ray, worker should only see one GPU or None
        device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        # Create Actor Critic instance
        self.actor = actor_factory(device)
        self.actor.to(device)

        # Create Algorithm instance
        self.algo = algo_factory(self.actor, device)

        # Define counters
        self.iter = 0

        if initial_weights: # if remote worker

            # Create train environments, define initial train states
            self.envs_train = train_envs_factory(device, index_worker)
            self.obs, self.rhs, self.done = self.actor.policy_initial_states(self.envs_train.reset())

            # Create test environments (if creation function available)
            self.envs_test = test_envs_factory(device, index_worker, mode="test")

            # Create Storage instance and set world initial state
            self.storage = storage_factory(device)
            self.update_every = self.algo.update_every or self.storage.max_size

            # Print worker information
            self.print_worker_info()
            self.set_weights(initial_weights)

            # Collect initial samples
            print("Collecting initial samples...")
            self.collect_data(self.algo.start_steps, self.algo.start_steps)

    def collect_data(self, num_steps=None, min_fraction=1.0):
        """
        Collect data from interactions with the environments.

        Parameters
        ----------
        num_steps : int
            Target number of train environment steps to take.
        min_steps : int
            Minimum number of train environment steps to stop at if other
            workers have finished their collection task.
        """

        if self.iter % (self.algo.num_epochs * self.algo.num_mini_batch) != 0:
            return

        if num_steps is None:
            num_steps = int(self.update_every)

        min_steps = int(num_steps * min_fraction)

        t = time.time()
        for step in range(num_steps):

            # Predict next action, next rnn hidden state and algo-specific outputs
            act, clip_act, rhs, algo_data = self.algo.acting_step(self.obs, self.rhs, self.done)

            # Interact with envs_vector with predicted action (clipped within action space)
            obs2, reward, done, infos = self.envs_train.step(clip_act)

            # Prepare transition dict
            transition = {"obs": self.obs, "rhs": rhs, "act": act, "rew": reward, "obs2": obs2, "done": done}
            transition.update(algo_data)

            # Store transition in buffer
            self.storage.insert(transition)

            # Update current world state
            self.obs, self.rhs, self.done = obs2, rhs, done

            # Record model version used to collect data
            self.storage.ac_version = self.iter

            if check_message("sample") == b"stop" and step >= min_steps:
                self.num_steps_reached = step
                break

        # Record and return metrics
        self.collect_time = time.time() - t
        self.collect_samples = num_steps * self.envs_train.num_envs

    def compute_gradients(self, batch):
        """
        Calculate actor gradients and average them with grads from
        other workers.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute algo loss.

        Returns
        -------
        info : dict
            Summary dict of relevant gradient-related information.
        """

        t = time.time()
        grads, info = self.algo.compute_gradients(batch, grads_to_cpu=False)
        compute_grads_t = time.time() - t

        ###### ALLREDUCE ######################################################

        t = time.time()
        if torch.cuda.is_available():
            for g in grads:
                torch.distributed.all_reduce(g, op=torch.distributed.ReduceOp.SUM)
        else:
            torch.distributed.all_reduce_coalesced(grads, op=torch.distributed.ReduceOp.SUM)

        for p in self.actor.parameters():
            if p.grad is not None:
                p.grad /= self.distributed_world_size
        avg_grads_t = time.time() - t

        #######################################################################

        info.update({"scheme/seconds_to/compute_grads": compute_grads_t})
        info.update({"scheme/seconds_to/avg_grads_t": avg_grads_t})

        return info

    def update_networks(self):
        """Update Actor Critic model"""
        self.algo.apply_gradients()

    def evaluate(self):
        """
        Test current actor version in self.envs_test.

        Returns
        -------
        mean_test_perf : float
            Average accumulated reward over all tested episodes.
        """

        completed_episodes = []
        obs = self.envs_test.reset()
        rewards = np.zeros(obs.shape[0])
        obs, rhs, done = self.actor.policy_initial_states(obs)

        while len(completed_episodes) < self.algo.num_test_episodes:
            # Predict next action and rnn hidden state
            act, clip_act, rhs, _ = self.algo.acting_step(obs, rhs, done, deterministic=True)

            # Interact with env with predicted action (clipped within action space)
            obs2, reward, done, _ = self.envs_test.step(clip_act)

            # Keep track of episode rewards and completed episodes
            rewards += reward.cpu().squeeze(-1).numpy()
            completed_episodes.extend(
                rewards[done.cpu().squeeze(-1).numpy() == 1.0].tolist())
            rewards[done.cpu().squeeze(-1).numpy() == 1.0] = 0.0

            obs = obs2

        return np.mean(completed_episodes)

    def step(self):
        """
        Perform logical learning step. Training proceeds interleaving collection
        of data samples with gradient computations and policy updates.

        Returns
        -------
        info : dict
            Summary dict with relevant step information.

        """

        # Collect data and prepare data batches
        if self.iter % (self.algo.num_epochs * self.algo.num_mini_batch) == 0:
            self.storage.before_update(self.actor, self.algo)
            self.batches = self.storage.generate_batches(
                self.algo.num_mini_batch, self.algo.mini_batch_size,
                self.algo.num_epochs, self.actor.is_recurrent)
            collect_time, collected_samples = self.collect_time, self.collect_samples
        else:
            collect_time, collected_samples = 0.0, 0

        # Get next batch
        batch = self.batches.__next__()

        # Compute gradients
        info = self.compute_gradients(batch)

        # Update model
        self.update_networks()

        # Add extra information to info dict
        info.update({"ac_update_num": self.iter})
        info.update({"collected_samples": collected_samples})
        info.update({"scheme/seconds_to/collect": collect_time})
        info.update({"scheme/metrics/collection_gradient_delay": self.iter - self.storage.ac_version})
        if self.iter == 0: info["collected_samples"] += self.algo.start_steps * self.envs_train.num_envs

        # Evaluate current network
        if self.iter % self.algo.test_every == 0:
            if self.envs_test and self.algo.num_test_episodes > 0:
                test_perf = self.evaluate()
                info.update({"scheme/metrics/test_performance": test_perf})

        # Update counter
        self.iter += 1

        return info

    def setup_torch_data_parallel(self, url, world_rank, world_size, backend):
        """Join a torch process group for distributed SGD."""
        torch.distributed.init_process_group(
            backend=backend,
            init_method=url,
            rank=world_rank,
            world_size=world_size)
        self.distributed_world_size = world_size

    def get_node_ip(self):
        """Returns the IP address of the current node."""
        return get_node_ip_address()

    def find_free_port(self):
        """Returns a free port on the current node."""
        return find_free_port()

    def set_weights(self, weights):
        """
        Update the worker actor version with provided weights.

        weights: dict of tensors
            Dict containing actor weights to be set.
        """
        self.algo.set_weights(weights["weights"])

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.algo.update_algo_parameter(parameter_name, new_parameter_value)


class CGUWorkerSet(WS):
    """
    Class to better handle the operations of ensembles of workers.

    Parameters
    ----------
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    fraction_workers : float
        Minimum fraction of the remote workers to complete the data collection
        task to stop collection phase.
    fraction_samples : float
        Minimum fraction of the target steps allowed in a data collection.
    num_workers : int
        Number of remote workers in the worker set.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword arguments of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    num_updates : int
        Times step() function has been called.
    fraction_workers : float
        Minimum fraction of the remote workers to complete the data collection
        task to stop collection phase.
    fraction_samples : float
        Minimum fraction of the target steps allowed in a data collection.
    """

    def __init__(self,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda x, y, c: None,
                 worker_remote_config=default_remote_config,
                 fraction_workers=0.8,
                 fraction_samples=0.5,
                 num_workers=1):

        self.worker_class = CGUWorker
        default_remote_config.update(worker_remote_config)
        self.remote_config = default_remote_config
        self.worker_params = {
            "algo_factory": algo_factory,
            "storage_factory": storage_factory,
            "test_envs_factory": test_envs_factory,
            "train_envs_factory": train_envs_factory,
            "actor_factory": actor_factory,
        }

        super(CGUWorkerSet, self).__init__(
            worker=self.worker_class,
            worker_params=self.worker_params,
            worker_remote_config=self.remote_config,
            num_workers=num_workers)

        self.num_updates = 0
        self.num_workers = num_workers
        self.fraction_samples = fraction_samples if self.num_workers > 1 else 1.0
        self.fraction_workers = fraction_workers if self.num_workers > 1 else 1.0

        # Check remote workers exist
        if len(self.remote_workers()) == 0:
            raise ValueError("""at least 1 grad worker required""")

        # Setup the distributed processes for gradient averaging
        ip = ray.get(self.remote_workers()[0].get_node_ip.remote())
        port = ray.get(self.remote_workers()[0].find_free_port.remote())
        address = "tcp://{ip}:{port}".format(ip=ip, port=port)
        ray.get([worker.setup_torch_data_parallel.remote(
            address, i, len(self.remote_workers()), "nccl")
                 for i, worker in enumerate(self.remote_workers())])

    def step(self):
        """
        Takes a logical optimization step.

        Returns
        -------
        info : dict
            Summary dict of relevant information about the update process.
        """

        # Start data collection in all workers
        broadcast_message("sample", b"start-continue")

        pending_samples = [e.collect_data.remote(
            min_fraction=self.fraction_samples) for e in self.remote_workers()]

        # Keep checking how many workers have finished until percent% are ready
        samples_ready, samples_not_ready = ray.wait(
            pending_samples, num_returns=len(pending_samples), timeout=0.5)
        while len(samples_ready) < (self.num_workers * self.fraction_workers):
            samples_ready, samples_not_ready = ray.wait(
                pending_samples, num_returns=len(pending_samples), timeout=0.5)

        # Send stop message to the workers
        broadcast_message("sample", b"stop")

        # Compute model updates
        results = ray.get([e.step.remote() for e in self.remote_workers()])

        # Merge worker results
        step_metrics = defaultdict(float)
        for info in results:
            info["scheme/metrics/gradient_update_delay"] = self.num_updates - info.pop("ac_update_num")
            for k, v in info.items(): step_metrics[k] += v

        # Update info dict
        info = {k: v / self.num_workers for k, v in step_metrics.items()}

        # Update counters
        self.num_updates += 1

        return info

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        for e in self.remote_workers():
            e.update_algo_parameter.remote(parameter_name, new_parameter_value)

    def save_model(self, fname):
        """
        Save current version of actor as a torch loadable checkpoint.

        Parameters
        ----------
        fname : str
            Filename given to the checkpoint.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        model_dict = ray.get(self.remote_workers()[0].get_weights.remote())
        torch.save(model_dict, fname + ".tmp")
        os.rename(fname + '.tmp', fname)
        save_name = fname + ".{}".format(self.num_updates)
        copy2(fname, save_name)
        return save_name

    def update_workers(self):
        """Return update workers"""
        return self


from .c_workers import CWorkerSet
from .g_workers import GWorkerSet
from .u_worker import UWorker


class Workers:
    """
    Class to containing and handling all scheme workers.

    Parameters
    ----------
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    num_col_workers : int
        Number of remote workers performing collection operations.
    col_worker_remote_config : dict
        Ray resource specs for the remote collection workers.
    num_grad_workers : int
        Number of remote workers performing collection operations.
    grad_worker_remote_config : dict
        Ray resource specs for the remote gradient computation workers.
    """
    def __init__(self,
                 num_col_workers,
                 num_grad_workers,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda x, y, c: None,
                 col_worker_remote_config={"num_cpus": 1, "num_gpus": 0.5},
                 grad_worker_remote_config={"num_cpus": 1, "num_gpus": 0.5}):

        col_workers_factory = CWorkerSet.worker_set_factory(
            num_workers=num_col_workers//num_grad_workers,
            test_envs_factory=test_envs_factory,
            train_envs_factory=train_envs_factory,
            actor_factory=actor_factory,
            worker_remote_config=col_worker_remote_config)
        grad_workers = GWorkerSet(
            algo_factory=algo_factory,
            storage_factory=storage_factory,
            actor_factory=actor_factory,
            collection_workers_factory=col_workers_factory,
            num_workers=num_grad_workers,
            worker_remote_config=grad_worker_remote_config)
        self._update_worker = UWorker(grad_workers)

    def update_workers(self):
        """Return local worker"""
        return self._update_worker


from ..scheme_da2cs.c_workers import CWorker
from ..base.worker_set import WorkerSet as WS
from ..base.worker import default_remote_config


class CWorkerSet(WS):
    """
    Class to better handle the operations of ensembles of workers.

    Parameters
    ----------
    initial_weights : ray object ID
        Initial model weights.
    algo_factory : func
        A function that creates an algorithm class.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    actor_factory : func
        A function that creates a policy.
    test_envs_factory : func
        A function to create test environments.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    num_workers : int
        Number of remote workers in the worker set.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword parameters of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    """
    def __init__(self,
                 initial_weights,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda x, y, c: None,
                 worker_remote_config=default_remote_config,
                 num_workers=1):

        self.worker_class = CWorker
        default_remote_config.update(worker_remote_config)
        self.remote_config = default_remote_config
        self.worker_params = {
            "initial_weights": initial_weights,
            "algo_factory": algo_factory,
            "storage_factory": storage_factory,
            "test_envs_factory": test_envs_factory,
            "train_envs_factory": train_envs_factory,
            "actor_factory": actor_factory,
        }

        super(CWorkerSet, self).__init__(
            worker=self.worker_class,
            worker_params=self.worker_params,
            worker_remote_config=self.remote_config,
            num_workers=num_workers,
            add_local_worker=False)

    @classmethod
    def worker_set_factory(cls,
                           actor_factory,
                           test_envs_factory,
                           train_envs_factory,
                           worker_remote_config=default_remote_config,
                           num_workers=1):
        """
        Returns a function to create new CWorkerSet instances.

        Parameters
        ----------
        train_envs_factory : func
            A function to create train environments.
        actor_factory : func
            A function that creates a policy.
        test_envs_factory : func
            A function to create test environments.
        worker_remote_config : dict
            Ray resource specs for the remote workers.
        num_workers : int
            Number of remote workers in the worker set.

        Returns
        -------
        create_algo_instance : func
            creates a new CWorkerSet class instance.
        """
        def create_worker_set_instance(
                initial_weights,
                algo_factory,
                storage_factory):
            return cls(num_workers=num_workers,
                       initial_weights=initial_weights,
                       worker_remote_config=worker_remote_config,
                       algo_factory=algo_factory,
                       actor_factory=actor_factory,
                       storage_factory=storage_factory,
                       test_envs_factory=test_envs_factory,
                       train_envs_factory=train_envs_factory)
        return create_worker_set_instance



from .off_policy.replay_buffer import ReplayBuffer
from .off_policy.hindsight_experience_replay_buffer import HindsightExperienceReplayBuffer

from .on_policy.on_policy_buffer import OnPolicyBuffer
from .on_policy.on_policy_gae_buffer import OnPolicyGAEBuffer
from .on_policy.on_policy_vtrace_buffer import OnPolicyVTraceBuffer





import os
import ray
import torch
from shutil import copy2
from collections import defaultdict
from ..utils import ray_get_and_free


class UWorker:
    """
    Update worker. Handles actor updates.

    This worker receives gradients from gradient workers and then updates the
    its actor model. Updated weights are asynchronously sent back
    to gradient workers.

    Parameters
    ----------
    grad_workers : WorkerSet
        Set of workers computing and sending gradients to the UWorker.

    Attributes
    ----------
    num_updates : int
        Number of times the actor model has been updated.
    grad_workers : WorkerSet
        Set of workers computing and sending gradients to the UWorker.
    """

    def __init__(self, grad_workers, broadcast_interval=None):

        self.num_updates = 0
        self.grad_workers = grad_workers
        self.broadcast_interval = broadcast_interval or len(self.grad_workers.remote_workers())

        # Check remote workers exist
        if len(self.grad_workers.remote_workers()) == 0:
            raise ValueError("""at least 1 grad worker required""")

    def sync_weights(self):
        """Synchronize gradient worker models with updater worker model"""
        weights = ray.put({
            "update": self.num_updates,
            "weights": self.grad_workers.local_worker().get_weights()})
        for e in self.grad_workers.remote_workers():
            e.set_weights.remote(weights)

    def step(self):
        """
        Takes a logical asynchronous optimization step.

        Returns
        -------
        info : dict
            Summary dict of relevant information about the update process.
        """

        # If first call, call for gradients from all workers
        if self.num_updates == 0:
            self.pending_gradients = {}
            for e in self.grad_workers.remote_workers():
                future = e.step.remote()
                self.pending_gradients[future] = e

        # Wait for first gradients ready
        wait_results = ray.wait(list(self.pending_gradients.keys()))
        ready_list = wait_results[0]
        future = ready_list[0]

        # Get gradients
        gradients, info = ray_get_and_free(future)

        # Update info dict
        info[
            "scheme/metrics/gradient_update_delay"] = self.num_updates - info.pop(
            "ac_update_num")

        # Update local worker weights
        self.grad_workers.local_worker().update_networks(gradients)
        e = self.pending_gradients.pop(future)

        # Update remote worker model version
        if self.num_updates % self.broadcast_interval == 0:
            weights = ray.put({
                "update": self.num_updates,
                "weights": self.grad_workers.local_worker().get_weights()})
            e.set_weights.remote(weights)

        # Call compute_gradients in remote worker again
        future = e.step.remote()
        self.pending_gradients[future] = e

        # Update counter
        self.num_updates += 1

        return info

    def step1(self):
        """
        Takes a logical asynchronous optimization step.

        Returns
        -------
        info : dict
            Summary dict of relevant information about the update process.
        """

        num_gradients = 0
        step_metrics = defaultdict(float)
        self.sync_weights()

        # Call for gradients from all workers
        pending_gradients = {}
        for e in self.grad_workers.remote_workers():
            future = e.step.remote()
            pending_gradients[future] = e

        while pending_gradients:

            # Wait for first gradients ready
            wait_results = ray.wait(list(pending_gradients.keys()))
            ready_list = wait_results[0]
            future = ready_list[0]

            # Get gradients
            gradients, info = ray_get_and_free(future)

            # Update local worker weights
            self.grad_workers.local_worker().update_networks(gradients)

            # Update info dict
            info["scheme/metrics/gradient_update_delay"] = self.num_updates - info.pop("ac_update_num")

            # Update counters
            for k, v in info.items(): step_metrics[k] += v

            if num_gradients < self.broadcast_interval:
                e = pending_gradients.pop(future)
                future = e.step.remote()
                pending_gradients[future] = e
                num_gradients += 1

        # Update counter
        self.num_updates += 1

        # Update info dict
        info = {k: v/self.broadcast_interval if k != "collected_samples"
        else v for k, v in step_metrics.items()}

        return info

    def save_model(self, fname):
        """
        Save current version of actor as a torch loadable checkpoint.

        Parameters
        ----------
        fname : str
            Filename given to the checkpoint.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        torch.save(self.grad_workers.local_worker().actor.state_dict(), fname + ".tmp")
        os.rename(fname + '.tmp', fname)
        save_name = fname + ".{}".format(self.num_updates)
        copy2(fname, save_name)
        return save_name

    def stop(self):
        """Stop remote workers"""
        for e in self.grad_workers.remote_workers():
            e.terminate_worker.remote()

    def adjust_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.grad_workers.local_worker().adjust_algo_parameter(parameter_name, new_parameter_value)
        for e in self.grad_workers.remote_workers():
            e.adjust_algo_parameter.remote(parameter_name, new_parameter_value)


import numpy as np
import torch

class HindsightExperienceReplayBuffer:
    """
    Storage class for Off-Policy algorithms using HER (https://arxiv.org/abs/1707.01495).

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    her_function : func
        Function to update obs, rhs, obs2 and rew according to HER paper.

    Attributes
    ----------
    off_policy_data_fields : tuple
        Accepted data fields. If the samples inserted contain other fields,
        an AssertionError will be raised.
    max_size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place.
    her_function : func
        Function to update obs, rhs, obs2 and rew according to HER paper.
    """

    off_policy_data_fields = ("obs", "obs2", "rhs", "act", "rew", "done")

    def __init__(self, size, device, her_function=lambda o, rhs, o2, r, goal_state : (o, rhs, o2, r)):

        self.device = device
        self.max_size = size
        self.her_function = her_function
        self.reset()

    @classmethod
    def create_factory(cls, size, her_function=lambda o, rhs, o2, r, goal_state : (o, rhs, o2, r)):
        """
        Returns a function that creates HindsightExperienceReplayBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.

        Returns
        -------
        create_buffer_instance : func
            creates a new OnPolicyBuffer class instance.
        """
        def create_buffer(device):
            """Create and return a ReplayBuffer instance."""
            return cls(size, device, her_function)
        return create_buffer

    def init_tensors(self, sample):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        assert set(sample.keys()) == set(self.data.keys())
        self.data["obs"] = np.zeros((self.max_size, *sample["obs"].shape), dtype=np.float32)
        self.data["obs2"] = np.zeros((self.max_size, *sample["obs"].shape), dtype=np.float32)
        self.data["rhs"] = np.zeros((self.max_size, *sample["rhs"].shape), dtype=np.float32)
        self.data["act"] = np.zeros((self.max_size, *sample["act"].shape), dtype=np.float32)
        self.data["rew"] = np.zeros((self.max_size, *sample["rew"].shape), dtype=np.float32)
        self.data["done"] = np.zeros((self.max_size, *sample["done"].shape), dtype=np.float32)

    def get_data(self):
        """Return currently stored data."""
        return {k: v[0:self.step] for k, v in self.data.items()}

    def reset(self):
        """Set class counters to zero and remove stored data"""
        self.step, self.size, self.last_episode_start = 0, 0, 0
        self.data = {k: None for k in self.off_policy_data_fields} # lazy init

    def add_data(self, new_data):
        """
        Appends new_data to currently stored data.

        Parameters
        ----------
        new_data : dict
            Dictionary of env transition samples to be added to self.data.
        """
        for k, v in new_data.items():

            if self.data[k] is None:
                self.data[k] = np.zeros((self.max_size, *v.shape[1:]), dtype=np.float32)

            len, num_proc = v.shape[0:2]

            if self.step + len <= self.max_size:
                self.data[k][self.step:self.step + len] = v
            else:
                self.data[k][self.step:self.max_size] = v[0:self.max_size - self.step]
                self.data[k][0:len - self.max_size - self.step] = v[self.max_size - self.step:]

        self.step = (self.step + len) % self.max_size
        self.size = min(self.size + len, self.max_size)

    def insert(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        if self.size == 0: # data tensors lazy initialization
            self.init_tensors(sample)

        # Add current observation
        self.data["obs"][self.step] = sample["obs"].cpu()
        self.data["rhs"][self.step] = sample["rhs"].cpu()
        self.data["act"][self.step] = sample["act"].cpu()
        self.data["rew"][self.step] = sample["rew"].cpu()
        self.data["obs2"][self.step] = sample["obs2"].cpu()
        self.data["done"][self.step] = sample["done"].cpu()

        self.step = (self.step + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

        # Handle end of episode - fixed episode length case.
        if (sample["done"].cpu() == 1.0).all():

            goal_state = sample["obs2"].cpu()

            for i in range(self.last_episode_start, self.step):

                self.data["act"][self.step] = sample["act"].cpu()
                self.data["done"][self.step] = sample["done"].cpu()

                obs, rhs, obs2, rew = self.her_function(
                    sample["obs"].cpu(),
                    sample["rhs"].cpu(),
                    sample["obs2"].cpu(),
                    sample["rew"].cpu(),
                    goal_state)

                self.data["obs"][self.step] = obs
                self.data["rhs"][self.step] = rhs
                self.data["rew"][self.step] = rew
                self.data["obs2"][self.step] = obs2

                self.step = (self.step + 1) % self.max_size
                self.size = min(self.size + 1, self.max_size)

            self.last_episode_start = self.step

    def before_update(self, actor, algo):
        """
        Steps required before updating actor policy model.

        Parameters
        ----------
        actor : ActorCritic
            An actor class instance.
        algo : an algorithm class
            An algorithm class instance.
        """
        pass

    def after_update(self):
        """Steps required after updating actor policy model"""
        pass

    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1,
                         recurrent_ac=False):
        """
        Returns a batch iterator to update actor.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.
        recurrent_ac : bool
            Whether actor policy is a RNN or not.
        shuffle : bool
            Whether to shuffle collected data or generate sequential

        Yields
        ______
        batch : dict
            Generated data batches.
        """
        num_proc = self.data["obs"].shape[1]

        if recurrent_ac:  # Batches for a feed recurrent actor
            raise NotImplementedError

        else:  # Batches for a feed forward actor
            for _ in range(num_mini_batch):
                idxs = np.random.randint(0, num_proc * self.size, size=mini_batch_size)
                batch = dict(
                    obs=self.data["obs"][0:self.size].reshape(-1, *self.data["obs"].shape[2:])[idxs],
                    rhs=self.data["rhs"][0:self.size].reshape(-1, *self.data["rhs"].shape[2:])[idxs],
                    act=self.data["act"][0:self.size].reshape(-1, *self.data["act"].shape[2:])[idxs],
                    rew=self.data["rew"][0:self.size].reshape(-1, *self.data["rew"].shape[2:])[idxs],
                    obs2=self.data["obs2"][0:self.size].reshape(-1, *self.data["obs2"].shape[2:])[idxs],
                    done=self.data["done"][0:self.size].reshape(-1, *self.data["done"].shape[2:])[idxs])
                yield {k: torch.as_tensor(v, dtype=torch.float32).to(self.device) for k, v in batch.items()}



import gym
from ..common import FrameStack, FrameSkip


def make_mujoco_train_env(env_id, index_worker=0, index_env=0, seed=0, frame_skip=0, frame_stack=1):
    """
    Create train MuJoCo environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.

    Returns
    -------
    env : gym.Env
        Train environment.
    """

    env = gym.make(env_id)
    env.seed(seed + index_worker + index_env)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    return env

def make_mujoco_test_env(env_id, index_worker=0, index_env=0, seed=0, frame_skip=0, frame_stack=1):
    """
    Create test MuJoCo environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.

    Returns
    -------
    env : gym.Env
        Test environment.
    """
    env = gym.make(env_id)
    env.seed(seed + index_worker + index_env)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    return env


import torch.nn as nn


def init(module, weight_init, bias_init, gain=1):
    """
    Parameters
    ----------
    module : nn.Module
        nn.Module to initialize.
    weight_init : func
        Function to initialize module weights.
    bias_init : func
        Function to initialize module biases.
    Returns
    -------
    module : nn.Module
        Initialized module
    """
    weight_init(module.weight.data, gain=gain)
    weight_init(module.weight.data)
    bias_init(module.bias.data)
    return module


import setuptools
import subprocess

try:
    version = (
        subprocess.check_output(["git", "describe", "--abbrev=0", "--tags"])
        .strip()
        .decode("utf-8")
    )
except Exception as e:
    print("Could not get version tag. Defaulting to version 0")
    version = "0"

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

if __name__ == "__main__":
    with open("README.md", "r") as fh:
        long_description = fh.read()

    setuptools.setup(
        name="nappo",
        version=version,
        author="",
        author_email="",
        description="Disributed RL implementations with ray and pytorch.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/nappo/nappo/",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        packages=setuptools.find_packages(include=["nappo*"], exclude=[]),
        install_requires=requirements,
    )


import ray
import time
import torch
import threading
from six.moves import queue
from functools import partial
from collections import defaultdict, deque
from ..utils import TaskPool, ray_get_and_free


class GUWorker:
    """
    Update worker. Handles actor updates.

    This class coordinates sequential central actor optimization, using
    rollouts collected by distributed workers to compute gradients and update
    the models.

    Parameters
    ----------
    workers : WorkerSet
        Set of workers collecting and sending rollouts to the UWorker.
    device : torch.device
        CPU or specific GPU to use for computation.
    broadcast_interval : int
        After how many central updates, model weights should be broadcasted to
        remote collection workers.
    updater_queue_size : int
        Maximum number of data dicts fitting in the updater queue.
    max_collect_requests_pending : int
        Maximum number of collection tasks simultaneously scheduled to each
        collection worker.

    Attributes
    ----------
    local_worker : Worker
        Local worker that acts as a parameter server.
    remote_workers : list of Workers
        Set of workers collecting and sending rollouts.
    num_updates : int
        Number of times the actor model has been updated.
    num_workers : int
        number of remote workers computing gradients.
    inqueue : queue.Queue
        Queue to store the data dicts received and pending to be processed.
    outqueue : queue.Queue
         Queue to store the info dicts resulting from the model update operation.
    """
    def __init__(self,
                 workers,
                 device="cpu",
                 broadcast_interval=1,
                 updater_queue_size=100,
                 max_collect_requests_pending=2):

        self.local_worker = workers.local_worker()
        self.local_worker.actor.to(device)
        self.remote_workers = workers.remote_workers()
        self.num_workers = len(self.remote_workers)

        # Check remote workers exist
        if self.num_workers == 0:
            raise ValueError("""At least 1 data collection worker required""")

        # Queues
        self.inqueue = queue.Queue(maxsize=updater_queue_size)
        self.outqueue = queue.Queue()

        # Counters and metrics
        self.num_sent_since_broadcast = 0
        self.local_worker.num_updates = 0
        self.metrics = defaultdict(partial(deque, maxlen=100))

        # Create CollectorThread
        self.collector = CollectorThread(
            input_queue=self.inqueue,
            local_worker=self.local_worker,
            remote_workers=self.remote_workers,
            broadcast_interval=broadcast_interval,
            max_collect_requests_pending=max_collect_requests_pending)

        # Start CollectorThread
        self.collector.start()

        # Create UpdaterThread
        self.updater = UpdaterThread(
            input_queue=self.inqueue,
            output_queue=self.outqueue,
            local_worker=self.local_worker)

        # Start UpdaterThread
        self.updater.start()

    @property
    def num_updates(self):
        return self.local_worker.num_updates

    def step(self):
        """Collect and returns information from executed training steps."""

        # Check results in parameter server output queue
        step_metrics = defaultdict(float)
        while not self.outqueue.empty():
            info = self.outqueue.get()
            for k, v in info.items(): step_metrics[k] += v

        # Update info dict
        info = {k: v / self.num_workers if k != "collected_samples"
        else v for k, v in step_metrics.items()}

        return info

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.local_worker.algo.update_algo_parameter(parameter_name, new_parameter_value)

    def stop(self):
        """Stop collecting data and updating the local policy."""
        self.collector.stop()
        self.updater.stop()


class CollectorThread(threading.Thread):
    """
    This class receives data from the workers and queues it to the updater queue.


    Parameters
    ----------
    inqueue : queue.Queue
        Queue to store the data dicts received and pending to be processed.
    local_worker : Worker
        Local worker that acts as a parameter server.
    remote_workers : list of Workers
        Set of workers collecting and sending rollouts.
    broadcast_interval : int
        After how many central updates, model weights should be broadcasted to
        remote collection workers.
    max_collect_requests_pending : int
        Maximum number of collection tasks simultaneously scheduled to each
        collection worker.

    Attributes
    ----------
    input_queue : queue.Queue
        Queue to store the data dicts received and pending to be processed.
    local_worker : Worker
        Local worker that acts as a parameter server.
    remote_workers : list of Workers
        Set of workers collecting and sending rollouts.
    broadcast_interval : int
        After how many central updates, model weights should be broadcasted to
        remote collection workers.
    num_sent_since_broadcast : int
        Number of data dicts received since last model weights were broadcasted.
    num_workers : int
        number of remote workers computing gradients.
    collector_tasks : TaskPool
        Task pool to track remote workers in-flight collection tasks.
    stopped : bool
        Whether or not the thread in running.
    """

    def __init__(self,
                 input_queue,
                 local_worker,
                 remote_workers,
                 broadcast_interval=1,
                 max_collect_requests_pending=2):

        threading.Thread.__init__(self)

        self.inqueue = input_queue
        self.local_worker = local_worker
        self.remote_workers = remote_workers
        self.broadcast_interval = broadcast_interval
        self.num_workers = len(self.remote_workers)

        # Counters and metrics
        self.num_sent_since_broadcast = 0
        self.metrics = defaultdict(partial(deque, maxlen=100))

        # Start collecting data
        self.collector_tasks = TaskPool()
        for ev in self.remote_workers:
            for _ in range(max_collect_requests_pending):
                self.collector_tasks.add(ev, ev.collect.remote())

        self.stopped = False

    def run(self):
        while not self.stopped:
            self.step()

    def step(self):
        """
        Continuously collects data from remote workers and puts it
        in the updater queue.
        """

        # Wait to remote workers to complete data collection tasks
        for e, rollouts in self.collector_tasks.completed(blocking_wait=False, max_yield=1):

            # Move new collected rollouts to parameter server input queue
            self.inqueue.put(ray_get_and_free(rollouts))

            # Update counter and broadcast weights to worker if necessary
            self.num_sent_since_broadcast += 1
            if self.should_broadcast():
                self.broadcast_new_weights()

            # Request more data from worker
            self.collector_tasks.add(e, e.collect.remote())

    def should_broadcast(self):
        """Returns whether broadcast() should be called to update weights."""
        return self.num_sent_since_broadcast >= self.broadcast_interval

    def broadcast_new_weights(self):
        """Broadcast a new set of weights from the local worker."""
        latest_weights = ray.put({
            "update": self.local_worker.num_updates,
            "weights": self.local_worker.get_weights()})
        for e in self.remote_workers:
            e.set_weights.remote(latest_weights)
        self.num_sent_since_broadcast = 0

    def stop(self):
        """Stop collecting data."""
        self.stopped = True
        for e in self.remote_workers.remote_workers():
            e.terminate_worker.remote()

class UpdaterThread(threading.Thread):
    """
    This class receives data from the workers and continuously updates central actor.

    Parameters
    ----------
    input_queue : queue.Queue
        Queue to store the data dicts received and pending to be processed.
    output_queue : queue.Queue
        Queue to store the info dicts resulting from the model update operation.
    local_worker : Worker
        Local worker that acts as a parameter server.

    Attributes
    ----------
    local_worker : Worker
        Local worker that acts as a parameter server.
    input_queue : queue.Queue
        Queue to store the data dicts received and pending to be processed.
    output_queue : queue.Queue
        Queue to store the info dicts resulting from the model update operation.
    stopped : bool
        Whether or not the thread in running.
    """

    def __init__(self,
                 input_queue,
                 output_queue,
                 local_worker):

        threading.Thread.__init__(self)

        self.stopped = False
        self.inqueue = input_queue
        self.outqueue = output_queue
        self.local_worker = local_worker

    def run(self):
        while not self.stopped:
            self.step()

    def compute_gradients(self, batch):
        """
        Calculate actor gradients.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute algo loss.

        Returns
        -------
        info : dict
            Summary dict of relevant gradient-related information.
        """
        t = time.time()
        _, info = self.local_worker.algo.compute_gradients(batch)
        info.update({"scheme/seconds_to/compute_grads_t": time.time() - t})
        return info

    def update_networks(self):
        """Update Actor Critic model"""
        self.local_worker.algo.apply_gradients()

    def step(self):
        """
        Continuously pulls data from the input queue, computes gradients,
        updates the local actor model and places information in the
        output queue.
        """

        n = self.local_worker.algo.num_epochs * self.local_worker.algo.num_mini_batch
        if self.local_worker.num_updates % n == 0:

            new_rollouts = self.inqueue.get(timeout=30)
            self.local_worker.storage.add_data(new_rollouts["data"])
            self.rollouts_info = new_rollouts["info"]
            self.local_worker.storage.before_update(
                self.local_worker.actor, self.local_worker.algo)

            # Prepare data batches
            self.batches = self.local_worker.storage.generate_batches(
                self.local_worker.algo.num_mini_batch, self.local_worker.algo.mini_batch_size,
                self.local_worker.algo.num_epochs, self.local_worker.actor.is_recurrent)

        # Compute grads
        info = self.compute_gradients(self.batches.__next__())

        # Apply grads
        self.update_networks()

        # Add extra information to info dict
        info.update(self.rollouts_info)
        info.update({"scheme/metrics/gradient_update_delay": 0})
        info.update({"scheme/metrics/collection_gradient_delay":
        self.local_worker.num_updates - self.rollouts_info["ac_version"]})

        # Update counter
        self.local_worker.num_updates += 1
        self.rollouts_info["collected_samples"] = 0  # count only once

        self.outqueue.put(info)

    def stop(self):
        """Stop updating the local policy."""
        self.stopped = True


import ray
import time
from collections import deque

FREE_DELAY_S = 10.0
MAX_FREE_QUEUE_SIZE = 100
_last_free_time = 0.0
_to_free = []

def ray_get_and_free(object_ids):
    """
    Call ray.get and then queue the object ids for deletion.
    This function should be used whenever possible in RLlib, to optimize
    memory usage. The only exception is when an object_id is shared among
    multiple readers.

    Adapted from https://github.com/ray-project/ray/blob/master/rllib/utils/memory.py

    Parameters
    ----------
    object_ids : ObjectID|List[ObjectID]
        Object ids to fetch and free.

    Returns
    -------
    result : python objects
        The result of ray.get(object_ids).
    """

    global _last_free_time
    global _to_free

    result = ray.get(object_ids)
    if type(object_ids) is not list:
        object_ids = [object_ids]
    _to_free.extend(object_ids)

    # batch calls to free to reduce overheads
    now = time.time()
    if (len(_to_free) > MAX_FREE_QUEUE_SIZE
        or now - _last_free_time > FREE_DELAY_S):
        ray.internal.free(_to_free)
        _to_free = []
        _last_free_time = now

    return result


class TaskPool:
    """
    Class to track the status of many in-flight actor tasks.

    Adapted from https://github.com/ray-project/ray/blob/master/rllib/utils/actors.py
    """

    def __init__(self):
        self._obj_task_to_worker = {}
        self.fetching = deque()

    def add(self, worker, obj_id):
        self._obj_task_to_worker[obj_id] = worker

    def completed(self, blocking_wait=False, max_yield=999):
        """ Yield completed task objects."""
        pending = list(self._obj_task_to_worker)
        if pending:
            ready, _ = ray.wait(pending, num_returns=max_yield if max_yield < len(pending) else len(pending), timeout=10)

            if not ready and blocking_wait:
                while not ready:
                    ready, _ = ray.wait(pending, num_returns=1, timeout=10.0)

            for obj_id in ready:
                yield self._obj_task_to_worker.pop(obj_id), obj_id

def average_gradients(grads_list):
    """
    Averages gradients coming from distributed workers.

    Parameters
    ----------
    grads_list : list of lists of tensors
        List of actor gradients from different workers.
    Returns
    -------
    avg_grads : list of tensors
        Averaged actor gradients.
    """
    avg_grads = [
        sum(d[grad] for d in grads_list) / len(grads_list) if
        grads_list[0][grad] is not None else 0.0
        for grad in range(len(grads_list[0]))]
    return avg_grads





import torch
from .on_policy_buffer import OnPolicyBuffer as B


class OnPolicyGAEBuffer(B):
    """
    Storage class for On-Policy algorithms with Generalized Advantage
    Estimator (GAE). https://arxiv.org/abs/1506.02438

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    gae_lambda : float
        GAE lambda parameter.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.

    Attributes
    ----------
    on_policy_data_fields : tuple
        Accepted data fields. If the samples inserted contain other fields,
        an AssertionError will be raised.
    max_size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place.
    gae_lambda : float
        GAE lambda parameter.
    """

    on_policy_data_fields = ("obs", "obs2", "rhs", "act", "rew", "val", "logp", "done")

    def __init__(self, size, gae_lambda=0.95, device=torch.device("cpu")):

        super(OnPolicyGAEBuffer, self).__init__(
            size=size,
            device=device)

        self.gae_lambda = gae_lambda

    @classmethod
    def create_factory(cls, size, gae_lambda=0.95):
        """
        Returns a function that creates OnPolicyGAEBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        gae_lambda : float
            GAE lambda parameter.

        Returns
        -------
        create_buffer_instance : func
            creates a new OnPolicyBuffer class instance.
        """
        def create_buffer_instance(device):
            """Create and return a OnPolicyGAEBuffer instance."""
            return cls(size, gae_lambda, device)
        return create_buffer_instance

    def before_update(self, actor, algo):
        """
        Before updating actor policy model, compute returns and advantages.

        Parameters
        ----------
        actor : ActorCritic
            An actor class instance.
        algo : an algorithm class
            An algorithm class instance.
        """
        with torch.no_grad():
            _ = actor.get_action(
                self.data["obs"][self.step - 1],
                self.data["rhs"][self.step - 1],
                self.data["done"][self.step - 1])
            next_value = actor.get_value(self.data["obs"][self.step - 1])

        self.data["val"][self.step] = next_value
        self.compute_returns(algo.gamma)
        self.compute_advantages()

    def compute_returns(self, gamma):
        """
        Compute return values.

        Parameters
        ----------
        gamma : float
            Algorithm discount factor parameter.
        """
        len = self.step if self.step != 0 else self.max_size
        gae = 0
        for step in reversed(range(len)):
            delta = (self.data["rew"][step] + gamma * self.data["val"][step + 1] * (
                1.0 - self.data["done"][step + 1]) - self.data["val"][step])
            gae = delta + gamma * self.gae_lambda * (1.0 - self.data["done"][step + 1]) * gae
            self.data["ret"][step] = gae + self.data["val"][step]



from .feature_extractors.mlp import MLP
from .feature_extractors.cnn import CNN
from .feature_extractors.fixup_cnn import FixupCNN
from .neural_network import NNBase


def get_model(name):
    """Returns model class from name."""
    if name == "MLP":
        return MLP
    elif name == "CNN":
        return CNN
    elif name == "Fixup":
        return FixupCNN
    else:
        raise ValueError("Specified model not found!")




import torch
import torch.nn as nn
from torch.distributions.normal import Normal


class Categorical(nn.Module):
    """
    Categorical action distribution.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of action in action space.

    Attributes
    ----------
    linear: nn.Module
        Maps the incoming feature maps to probabilities over actions in output space.
    """
    def __init__(self, num_inputs, num_outputs):
        super(Categorical, self).__init__()
        self.linear = nn.Linear(num_inputs, num_outputs)

    def forward(self, x, deterministic=False):
        """
        Predict distribution parameters from x (obs features) and return
        actions (sampled and clipped), sampled action log
        probability and distribution entropy.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        logp_action : torch.tensor
            Log probability of `action` according to the predicted action distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        """

        # Predict distribution parameters
        x = self.linear(x)

        # Create distribution and sample action
        action_dist = torch.distributions.Categorical(logits=x)
        self.action_dist = action_dist # ugly hack to handle sac discrete case

        if deterministic:
            action = clipped_action = action_dist.probs.argmax(dim=-1, keepdim=True)
        else:
            action = clipped_action = action_dist.sample().unsqueeze(-1)

        # Action log probability
        logp_action = action_dist.log_prob(action.squeeze( -1)).unsqueeze(-1)

        # Distribution entropy
        entropy_dist = action_dist.entropy().mean()

        return action, clipped_action, logp_action, entropy_dist


    def evaluate_actions(self, x, action):
        """
        Return log prob of action under the distribution generated from
        x (obs features). Also return entropy of the generated distribution.

        Parameters
        ----------
        x : torch.tensor
            obs feature map obtained from a policy_net.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted.
        """

        # Predict distribution parameters
        x = self.linear(x)

        # Create distribution
        action_dist = torch.distributions.Categorical(logits=x)

        # Evaluate log prob of action under action_dist
        logp_action = action_dist.log_prob(action.squeeze(-1)).unsqueeze(-1).sum(-1, keepdim=True)

        # Distribution entropy
        entropy_dist = action_dist.entropy().mean()

        return logp_action, entropy_dist